/*******************************************************************
  uLan Communication - low level and link driver

  ul_usb.c	- driver implementation for usb devices

  (C) Copyright 1996,1997,2002 by Pavel Pisa 

  The uLan driver is distributed under the Gnu General Public Licence. 
  See file COPYING for details.

  Author reserves right to use and publish sources for embedded 
  applications under different conditions too.
 *******************************************************************/

#ifdef FOR_USER_SPACE 
  #include <stdio.h>
  #include <unistd.h>
  #include <time.h>
  #include <errno.h>
  #include <asm/io.h>
  #include <asm/bitops.h>
  #include <asm/atomic.h>
  #include <sys/perm.h>
  #include <string.h>
  #include "emu_irq.h"
#elif defined(_WIN32)
  #ifndef FOR_WIN_WDM
    #define FOR_WIN_KMD
    /*#define UL_WITH_UARTS*/
    /*#define UL_WITH_DEVFS*/
    /*#define UL_WITH_PCI*/
    #define UL_WITH_USB
    #include "ntddk.h"
  #else 
    #define DBG 1
    #define ULD_DEBUG_DEFAULT (0x01|0x08|0x10|0x40|0x80)
    /*#define UL_WITH_UARTS*/
    /*#define UL_WITH_DEVFS*/
    /*#define UL_WITH_PCI*/
    #define UL_WITH_USB
    #include "wdm.h"
    #include "ul_wdinc.h"
  #endif
#elif defined(__DJGPP__)
  int uld_djgpp_data_beg;
  void uld_djgpp_code_beg(void) {;};
  #include <unistd.h>
  #include <errno.h>
  #include <stdio.h>
  #include <time.h>
  #include <string.h>
#else
  #include <linux/config.h>
  #include <linux/module.h>
  #include <linux/version.h>
  #include <linux/errno.h>
  #include <linux/kernel.h>
  #include <linux/sched.h>
  #include <linux/ioport.h>
  #include <linux/sys.h>
  #include <linux/fs.h>
  #include <linux/interrupt.h>
  #include <linux/poll.h>
  #include <linux/delay.h>
  #include <asm/bitops.h>
  #include <asm/segment.h>
  #include <asm/system.h>
  #include <asm/io.h>
  #include <asm/irq.h>

  #ifdef __KERNEL__
    #ifndef KERNEL
      #define KERNEL
    #endif
  #endif

  #include "k_compat.h"

  #if (LINUX_VERSION_CODE < VERSION(2,5,50))
    #include <linux/modversions.h>
  #endif /* 2.5.50 */

  #if (LINUX_VERSION_CODE < VERSION(2,5,0))
    #include <linux/malloc.h>
  #else
    #include <linux/slab.h>
  #endif /* 2.5.0 */

  #if (LINUX_VERSION_CODE < VERSION(2,3,20))
    #include <asm/spinlock.h>
  #else
    #include <linux/spinlock.h>
  #endif

  #if (LINUX_VERSION_CODE >= VERSION(2,4,0))
    /*#define UL_WITH_UARTS*/
    #define UL_WITH_DEVFS    
    /*#define UL_WITH_PCI*/
    #define UL_WITH_USB
    #include <linux/miscdevice.h>
    #include <linux/init.h>
  #endif 

  #if (LINUX_VERSION_CODE < VERSION(2,5,40))
    #include <linux/tqueue.h>
  #else /* 2.5.40 */
    #include <linux/devfs_fs_kernel.h>
  #endif /* 2.5.40 */

  #ifdef UL_WITH_PCI
    #include <linux/pci.h>
  #endif

  #ifdef UL_WITH_USB
    #include <linux/usb.h>
    #include <linux/smp_lock.h>
  #endif

#endif

#include "ul_drv.h"
#include "ul_hdep.h"

#ifndef UL_DRV_IN_LIB
unsigned uld_debug_flg=ULD_DEBUG_DEFAULT;
#else
unsigned uld_debug_flg; /* Left application  set defaults */
#endif
int ulbuffer=0x8000;	/* default communication buffer size */

#ifdef UL_IRQ_LOCK_GINI
 UL_IRQ_LOCK_GINI	/* Inicialize global uLan spinlock */
#endif

#ifdef UL_GLOBAL_IRQ_LOCK_GINI
 UL_GLOBAL_IRQ_LOCK_GINI /* Initialize global IRQ lock,
 			    very bad if needed */
#endif

#ifdef ENABLE_UL_MEM_CHECK
atomic_t ul_mem_check_counter=ATOMIC_INIT(0);
#endif /* ENABLE_UL_MEM_CHECK */

int uld_null_fnc(ul_drv *udrv, int ret_code)
{
  return 0;
}

/*******************************************************************/
/* Data iterator */

long ul_di_seek(ul_data_it *di,long pos, int whence)
{
  switch(whence)
  {
    case 1: pos+=di->pos; break;
    case 2: pos+=UL_BLK_HEAD(di->head_blk).len; break;
  };
  if(pos<di->pos)
  {
    di->blk=di->head_blk;
    di->pos=0;
    di->ptr=UL_BLK_HEAD(di->head_blk).fdata;
    if(pos<0) return 0;
  };
  di->ptr+=pos-di->pos;
  di->pos=pos;
  return pos;
};

uchar ul_di_read1(ul_data_it *di)
{
  uchar ch;
  if(ul_di_adjust(di))
    ch=*ul_di_byte(di);
  else ch=0;
  ul_di_inc(di);
  return ch;
};

uchar ul_di_write1(ul_data_it *di, uchar ch)
{
  ul_mem_blk *blk;
  while(!ul_di_adjust(di))
  {
    if(!(blk=ul_alloc_blk(UL_BLK_HEAD(di->head_blk).bll->udrv))) return 0;
    di->blk->next=blk;
  };
  *ul_di_byte(di)=ch;
  ul_di_inc(di);
  return 1;
};

int ul_di_read(ul_data_it *di, char *buf, int count)
{
  int cn;
  int len;

  cn=count;
  while(cn>0)
  {
    if(!ul_di_adjust(di))
    {
      memset(buf,0,cn);
      cn=0;
      break;
    };
    len=ul_di_atonce(di);
    if(len>cn) len=cn;
    memcpy(buf,ul_di_byte(di),len);
    ul_di_add(di,len);
    buf+=len;
    cn-=len;
  };
  return count-cn;
}

int ul_di_write(ul_data_it *di, const char *buf, int count)
{
  int cn;
  int len;
  ul_mem_blk *blk;

  cn=count;
  while(cn>0)
  {
    while(!ul_di_adjust(di))
    {
      if(!(blk=ul_alloc_blk(UL_BLK_HEAD(di->head_blk).bll->udrv))) 
	{count-=cn; break;};
      memset(UL_BLK_NDATA(blk),0,UL_BLK_SIZE);
      di->blk->next=blk;
    };
    len=ul_di_atonce(di);
    if(len>cn) len=cn;
    memcpy(ul_di_byte(di),buf,len);
    ul_di_add(di,len);
    buf+=len;
    cn-=len;
  };
  return count;
}

/*******************************************************************/
/* uLan memory management and bidirection linked lists */

int ul_mem_init(ul_drv *udrv, int req_size)
{
  void **pptr;
  ul_mem_blk *bptr;
  pptr=(void **)(udrv->mem_ptr=MALLOC(req_size+2*sizeof(void*)));
  if(!pptr) return -1;
  *(pptr++)=0;
  *(pptr++)=(void *)req_size;
  udrv->free_blk=bptr=(ul_mem_blk *)pptr;
  udrv->free_blk_cnt=1;

  while((req_size-=sizeof(ul_mem_blk))>=sizeof(ul_mem_blk))
  {
    bptr->next=bptr+1;
    bptr++;
    udrv->free_blk_cnt++;
  }; 
  bptr->next=NULL;
  udrv->prep_bll.first=udrv->prep_bll.last=NULL;
  udrv->prep_bll.cnt=0;
  udrv->prep_bll.udrv=udrv; 
  udrv->work_bll.first=udrv->work_bll.last=NULL;
  udrv->work_bll.cnt=0;
  udrv->work_bll.udrv=udrv; 
  udrv->proc_bll.first=udrv->proc_bll.last=NULL;
  udrv->proc_bll.cnt=0;
  udrv->proc_bll.udrv=udrv;
  udrv->opan_bll.first=udrv->opan_bll.last=NULL;
  udrv->opan_bll.cnt=0;
  udrv->opan_bll.udrv=udrv;
  udrv->con_message=NULL;

  return 0;
};

int ul_mem_done(ul_drv *udrv)
{
  void *ptr;
  if(udrv->mem_ptr)
  { ptr=*(void **)udrv->mem_ptr;
    FREE(udrv->mem_ptr);
    udrv->mem_ptr=0;
  };
  return 0;
};

/*** Insert frame blk on list bll ***/
void ul_bll_ins(ul_blk_bll *bll, ul_mem_blk *blk)
{
  UL_DRV_LOCK_FINI
  /* ul_drv *udrv=bll->udrv; */
  UL_BLK_HEAD(blk).bll=bll;
  /* UL_BLK_HEAD(blk).next=NULL; */
  /* UL_BLK_HEAD(blk).ref_cnt=1; */
  /* UL_BLK_HEAD(blk).flg|=UL_BFL_LOCK; */
  UL_DRV_LOCK;
  if((UL_BLK_HEAD(blk).prev=bll->last)!=NULL)
    UL_BLK_HEAD(bll->last).next=blk;
  else bll->first=blk;
  bll->last=blk;
  bll->cnt++;
  UL_DRV_UNLOCK;
};

/*** Remove frame blk from its bll ***/
void ul_bll_del(ul_mem_blk *blk)
{
  UL_DRV_LOCK_FINI
  ul_blk_bll *bll;
  ul_mem_blk *p;
  bll=UL_BLK_HEAD(blk).bll;
  UL_DRV_LOCK;
  bll->cnt--;

  if((p=UL_BLK_HEAD(blk).next)!=NULL) 
  { UL_BLK_HEAD(p).prev=UL_BLK_HEAD(blk).prev;
    UL_BLK_HEAD(blk).next=NULL;
  }else bll->last=UL_BLK_HEAD(blk).prev;

  if((p=UL_BLK_HEAD(blk).prev)!=NULL) 
  { UL_BLK_HEAD(p).next=UL_BLK_HEAD(blk).next;
    UL_BLK_HEAD(blk).prev=NULL;
  }else bll->first=UL_BLK_HEAD(blk).next;

  UL_BLK_HEAD(blk).bll=NULL;
  UL_DRV_UNLOCK;
};

/*** Move one message starting at beg_blk to list new_bll ***/
void ul_bll_move_mes(ul_blk_bll *new_bll, ul_mem_blk *beg_blk)
{
  UL_DRV_LOCK_FINI
  /* ul_drv *udrv=bll->udrv; */
  ul_mem_blk *end_blk;
  int         cnt_blk=1;
  ul_blk_bll *old_bll;
  ul_mem_blk *p;

  if(beg_blk==NULL) return;
  end_blk=beg_blk;
  while((UL_BLK_HEAD(end_blk).flg&UL_BFL_TAIL)&&UL_BLK_HEAD(end_blk).next)
    { end_blk=UL_BLK_HEAD(end_blk).next; cnt_blk++; };
  UL_BLK_HEAD(end_blk).flg&=~UL_BFL_TAIL;

  UL_DRV_LOCK;
  if((old_bll=UL_BLK_HEAD(beg_blk).bll)!=NULL)
  {
    old_bll->cnt-=cnt_blk;
    if((p=UL_BLK_HEAD(beg_blk).prev)!=NULL)
    { UL_BLK_HEAD(p).next=UL_BLK_HEAD(end_blk).next;
      UL_BLK_HEAD(beg_blk).prev=NULL;
    } else old_bll->first=UL_BLK_HEAD(end_blk).next;
    if((p=UL_BLK_HEAD(end_blk).next)!=NULL)
    { UL_BLK_HEAD(p).prev=UL_BLK_HEAD(beg_blk).prev;
      UL_BLK_HEAD(end_blk).next=NULL;
    } else old_bll->last=UL_BLK_HEAD(beg_blk).prev;
  };
  UL_DRV_UNLOCK;

  p=beg_blk;
  do UL_BLK_HEAD(p).bll=new_bll; 
  while ((p=UL_BLK_HEAD(p).next)!=NULL);

  if(new_bll)
  {
    UL_DRV_LOCK;
    if((UL_BLK_HEAD(beg_blk).prev=new_bll->last)!=NULL)
      UL_BLK_HEAD(new_bll->last).next=beg_blk;
    else new_bll->first=beg_blk;
    new_bll->last=end_blk;
    new_bll->cnt+=cnt_blk;
    UL_DRV_UNLOCK;
  };
};

void ul_free_mes(ul_drv *udrv,ul_mem_blk *beg_blk)
{
  ul_mem_blk *blk, *next_blk;
  if(beg_blk==NULL) return;
  ul_bll_move_mes(NULL,beg_blk);
  do {
    next_blk=beg_blk;
    beg_blk=UL_BLK_HEAD(beg_blk).next;
    do {
      blk=next_blk;
      next_blk=blk->next;
      ul_free_blk(udrv,blk);
    } while(next_blk!=NULL);
  } while(beg_blk!=NULL);
};

/*** Free one message starting at beg_blk ***/
INLINE void ul_bll_free_mes(ul_mem_blk *beg_blk)
{
  if(beg_blk==NULL) return;
  ul_free_mes(UL_BLK_HEAD(beg_blk).bll->udrv,beg_blk);
};

ul_mem_blk *ul_new_frame_head(ul_drv *udrv, uchar dadr,
                         uchar sadr, uchar cmd,  unsigned flg)
{ ul_mem_blk *blk;
  blk=ul_alloc_blk(udrv);
  if(blk==NULL) return NULL;
  UL_BLK_HEAD(blk).prev=NULL;
  UL_BLK_HEAD(blk).next=NULL;
  UL_BLK_HEAD(blk).bll=NULL;
  UL_BLK_HEAD(blk).retry_cnt=0;
  UL_BLK_HEAD(blk).ref_cnt=0;
  UL_BLK_HEAD(blk).flg=flg;
  UL_BLK_HEAD(blk).len=0;
  UL_BLK_HEAD(blk).cmd=cmd;
  UL_BLK_HEAD(blk).dadr=dadr;
  UL_BLK_HEAD(blk).sadr=sadr;
  UL_BLK_HEAD(blk).stamp=0;
  return blk;
};

void ul_tail_frame_head(ul_mem_blk *beg_blk, ul_mem_blk *blk)
{
  while(UL_BLK_HEAD(beg_blk).next!=NULL) 
    beg_blk=UL_BLK_HEAD(beg_blk).next;
  UL_BLK_HEAD(beg_blk).next=blk;
  UL_BLK_HEAD(blk).prev=beg_blk;
  UL_BLK_HEAD(beg_blk).flg|=UL_BFL_TAIL;
};

unsigned ul_gen_stamp(void)
{
  UL_DRV_LOCK_FINI
  static unsigned ul_stamp_cnt=0; 
  unsigned stamp;
  UL_DRV_LOCK;
  ul_stamp_cnt++;
  ul_stamp_cnt&=0x7fffffff;
  if(!ul_stamp_cnt)ul_stamp_cnt++;
  stamp=ul_stamp_cnt;
  UL_DRV_UNLOCK;
  return stamp; 
};

int ul_inc_ref_cnt(ul_mem_blk *mes)
{
  int cnt;
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  UL_BLK_HEAD(mes).ref_cnt++;
  cnt=UL_BLK_HEAD(mes).ref_cnt;
  UL_DRV_UNLOCK;
  return cnt;
};

int ul_dec_ref_cnt(ul_mem_blk *mes)
{
  int cnt;
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  if(UL_BLK_HEAD(mes).ref_cnt)UL_BLK_HEAD(mes).ref_cnt--;
  cnt=UL_BLK_HEAD(mes).ref_cnt;
  UL_DRV_UNLOCK;
  if(!cnt)ul_bll_free_mes(mes);
  return cnt;
};

/*******************************************************************/

int ul_max_cycle_cnt=0;
int ul_max_cycle_time=0;
int ul_irq_seq_num=0;
#ifdef ENABLE_UL_IRQ_STALE_WDG	
int ul_irq_stale_wdg_cnt=0;
#endif /* ENABLE_UL_IRQ_STALE_WDG */

void ulan_do_wd_timer(unsigned long data)
{
  
}

void ulan_stroke(ul_drv *udrv)
{
  UL_DRV_LOCK_FINI
  if(!uld_test_dfl(udrv,CHIPOK)) return;
  if(udrv->prep_bll.first==NULL) return;
  if(udrv->fnc_stroke)
  { udrv->fnc_stroke(udrv);
    return;
  }
  UL_DRV_LOCK;
  #ifndef _WIN32
  { int delay;
    delay=udrv->last_ctrl&0x80?0:ULD_HZ/100+2;
    SCHEDULE_UDRV_WDTIM(udrv,jiffies+delay);
  }
  #else /* _WIN32 */
  { LARGE_INTEGER delay;
    delay=udrv->last_ctrl&0x80?RtlConvertLongToLargeInteger(-1):
		 	      RtlConvertLongToLargeInteger(-40*10000);
    KeSetTimer(&udrv->wd_timer,delay,&udrv->wd_timer_dpc);
  } 
  #endif /* _WIN32 */
  UL_DRV_UNLOCK;
};

/*******************************************************************/
/* initialize and free of one driver */ 

ul_chip_type_ent ul_chip_types[12]={
    #ifdef UL_WITH_UARTS
	{"82510",u510_init,UL_CHIPT_RQPORT,1,0},
	{"16450",u450_init,UL_CHIPT_RQPORT,1,0},
    #endif /*UL_WITH_UARTS*/
    #ifdef UL_WITH_PCI
	{"16950-adv",u950pci_init,UL_CHIPT_PCI|UL_CHIPT_NOPROBE,4,
		0x16954000|U950PCI_CHOPT_TXDTRNEG|U950PCI_CHOPT_HSPDOSC},
	{"16950-pci",u950pci_init,UL_CHIPT_PCI|UL_CHIPT_NOPROBE,4,
		0x16954000|U950PCI_CHOPT_RXDONRI},
    #endif  /* UL_WITH_PCI */
	{NULL,NULL},
};

/* set initial state */
void ul_drv_new_init_state(ul_drv* udrv, int my_adr)
{
  /* init sequencer */
  udrv->fnc_sp=&udrv->fnc_stack[0];
 #if defined(UL_WITH_UARTS) || defined(UL_WITH_PCI)
  udrv->fnc_act=&uld_drvloop;
 #else
  udrv->fnc_act=&uld_null_fnc;
 #endif
  udrv->my_adr=my_adr;
 #ifdef FOR_LINUX_KERNEL
  init_waitqueue_head(&udrv->wq_kwt);
 #endif
  udrv->fnc_timeout=NULL;
  /* udrv->flags=0; */
  udrv->magic=UL_DRV_MAGIC;
  udrv->operators=NULL;
}

/* setups buffers, ports and irq for sucesfully detected device */
int ul_drv_new_start(ul_drv* udrv, unsigned buffer_size)
{ int ret;
  int test_cnt;
  unsigned long probe_irqs;
  /* init ports */
  ret=udrv->fnc_cctrl(udrv,UL_CC_PINIT,0);
  UL_PRINTF(KERN_INFO "uLan_pinit ret %d\n",ret);
  if(ret>=0)
  { /* take buffer memory */
    ret=ul_mem_init(udrv,buffer_size);
    UL_PRINTF(KERN_INFO "ul_mem_init ret %d\n",ret);
    if(ret>=0)
    { /* request irq */
      test_cnt=3;
      ret=udrv->fnc_cctrl(udrv,UL_CC_RQIRQ,0);
     #if defined(UL_WITH_PCI) || defined(UL_WITH_UARTS)
      if(ret==UL_RC_ENOFNC){ /* default IRQ handling */
	ret=0;
	#ifdef FOR_LINUX_KERNEL
	while(test_cnt--&&udrv->irq<=0)
	{ /* probe for IRQ */
	  probe_irq_off(probe_irq_on());
	  probe_irqs=probe_irq_on();
	  ret=udrv->fnc_cctrl(udrv,UL_CC_GENIRQ,1);
	  udelay(100);
	  udrv->irq=probe_irq_off(probe_irqs);
	  ret=udrv->fnc_cctrl(udrv,UL_CC_GENIRQ,0);
	  UL_PRINTF(KERN_INFO "uLan ul_drv_new : probed irq %d\n",udrv->irq);
	};
	#endif /* FOR_LINUX_KERNEL */
	if(udrv->irq<=0) ret=-1;
	  #ifndef _WIN32
	  else ret=request_irq(udrv->irq, uld_irq_handler, 
                              UL_SA_IRQ_FLAGS , "ulan", udrv);
	  #endif /* _WIN32 */
      }
     #endif /*defined(UL_WITH_PCI) || defined(UL_WITH_UARTS)*/
      if (ret>=0)
      { /* sucessful initialization */
	INIT_UDRV_WDTIM(udrv);
	INIT_UDRV_BH(udrv);
	uld_set_dfl(udrv,CHIPOK);
	return 0;
      } else UL_PRINTF(KERN_ERR "uLan ul_drv_new : cannot request irq %d\n",udrv->irq);
      ul_mem_done(udrv);
    } else UL_PRINTF(KERN_ERR "uLan ul_drv_new : no memory for buffers\n");
    ret=udrv->fnc_cctrl(udrv,UL_CC_PDONE,0);
  } else UL_PRINTF(KERN_ERR "uLan ul_drv_new : port 0x%lx init error\n",
                   udrv->port);
  return -1;
}

#ifndef _WIN32

/* initialize new driver from given parameters */
ul_drv *ul_drv_new(int port, int irq, int baud, int my_adr, char *chip_name)
{
  ul_drv* udrv;
  int ret;
  int chipidx=0;
  int chipprobe=1;
  /* covert chip name to index */
  if(chip_name&&strcmp(chip_name,"auto")) {
    for(chipidx=0;1;chipidx++){
      if(!ul_chip_types[chipidx].chip_name){
	UL_PRINTF(KERN_ERR "uLan ul_drv_new : unknown chip type %s\n",chip_name);
	return NULL;
      }
      if(!strcmp(ul_chip_types[chipidx].chip_name,chip_name)){
	chipprobe=0; break;
      }
    }
  }
  UL_PRINTF(KERN_INFO "uLan ul_drv_new : chip %s index %d probe %d\n",
 		 chip_name?chip_name:"auto",chipidx,chipprobe);
  /* mem for ul_drv */
  if(!(udrv=MALLOC(sizeof(ul_drv)))) return NULL;
  /* clear memory */
  memset(udrv,0,sizeof(ul_drv));
  /* set initial state */
  ul_drv_new_init_state(udrv, my_adr);
  /* init chip driver */
  ret=-1;
  do{
    ul_chip_init_fnc *chip_init;
    int chip_flags=ul_chip_types[chipidx].chip_flags;
    int chip_options=ul_chip_types[chipidx].chip_options;
    chip_init=ul_chip_types[chipidx].chip_init;
    if(chipprobe&&(ul_chip_types[chipidx].chip_flags&UL_CHIPT_NOPROBE))
      continue;
    /* if(ret=u510_init(udrv, port, irq, baud))>=0) */
    if((chip_flags&UL_CHIPT_RQPORT)&&!port) continue;
    if((chip_flags&UL_CHIPT_RQIRQ)&&!irq) continue;
    if((ret=(*chip_init)(udrv, port, irq, baud, chip_options))>=0){
      UL_PRINTF(KERN_INFO "uLan ul_drv_new : %s found at 0x%lx\n",
   		 ul_chip_types[chipidx].chip_name,udrv->port);
      break;
    }
  }while(chipprobe&&ul_chip_types[++chipidx].chip_name);
  if(ret>=0)
  { /* setups buffers, ports and irq for sucesfully detected device */
    if(ul_drv_new_start(udrv,ulbuffer)>=0)
      return udrv;
  }
  UL_PRINTF(KERN_ERR "uLan ul_drv_new : failed to init\n");
  FREE(udrv);
  return NULL;
};

/* destroy driver */
void ul_drv_free(ul_drv *udrv)
{
  if(!udrv) return;
  if(uld_test_dfl(udrv,CHIPOK))
  { if(udrv->fnc_cctrl(udrv,UL_CC_FREEIRQ,0)==UL_RC_ENOFNC)
      if(udrv->irq)
        free_irq(udrv->irq,udrv); 
  }
  if(udrv->irq>0) ul_synchronize_irq(udrv->irq);
  KILL_UDRV_BH(udrv);
  DONE_UDRV_WDTIM(udrv);
  schedule();
  if(uld_test_dfl(udrv,CHIPOK))
  { uld_clear_dfl(udrv,CHIPOK);
    udrv->fnc_cctrl(udrv,UL_CC_PDONE,0);
  };
  if(udrv->irq>0) ul_synchronize_irq(udrv->irq);
  KILL_UDRV_BH(udrv);
  { /* first not complete sanity check of client list */
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    while((opptr=udrv->operators)!=NULL){
      udrv->operators=opptr->opnext; 
      if(udrv->operators!=NULL) udrv->operators->opprew=NULL;
      opptr->udrv=NULL;
     #ifdef FOR_LINUX_KERNEL
      wake_up_interruptible(&opptr->wqrec);
     #endif /*FOR_LINUX_KERNEL*/
    }
    UL_DRV_UNLOCK;
  };
  ul_mem_done(udrv);
  FREE(udrv);
};

#endif /* _WIN32 */

/*******************************************************************/
/* Some internal test routines */

#undef UL_INT_TST

#ifdef UL_INT_TST

/* standard command codes 
   00H .. 3FH    store to buffer 
   40H .. 7FH    store to buffer without ACK
   80H .. 9FH    proccess at onece
   A0H .. BFH    process with additional receive
   C0H .. FFH    process with additional send
*/

#define UL_CMD_RES	0x80	/* Reinitialize RS485 */
#define UL_CMD_SFT	0x81	/* Test free space in input buffer */
#define UL_CMD_SID	0xF0	/* Send identification */
#define UL_CMD_SFI	0xF1	/* Send amount of free space in IB */
#define UL_CMD_TF0	0x98	/* End of stepping */
#define UL_CMD_TF1	0x99	/* Begin of stepping */
#define UL_CMD_STP	0x9A	/* Do step */
#define UL_CMD_DEB	0x9B	/* Additional debug commands */
#define UL_CMD_SPC	0xDA	/* Send state - for 8051 PCL PCH PSW ACC */

#define UL_CMD_RDM	0xF8	/* Read memory   T T B B L L */
#define UL_CMD_WRM	0xB8	/* Write mamory  T T B B L L */

void genrdmemblk(ul_drv *udrv, uchar dadr, int mtyp, int mbeg, int mlen)
{
  ul_mem_blk *blk, *beg_blk;
  ul_data_it di;
  blk=ul_new_frame_head(udrv, 3, 2, UL_CMD_RDM,
       UL_BFL_M2IN|UL_BFL_TAIL|UL_BFL_SND|
       UL_BFL_PRQ|UL_BFL_ARQ);
  if(blk==NULL) return;
  beg_blk=blk;
  ul_di_init(&di,blk);
  /* ul_di_write1(&di,0); */
  ul_di_write1(&di,(uchar)mtyp);
  ul_di_write1(&di,(uchar)(mtyp>>8));
  ul_di_write1(&di,(uchar)mbeg);
  ul_di_write1(&di,(uchar)(mbeg>>8));
  ul_di_write1(&di,(uchar)mlen);
  ul_di_write1(&di,(uchar)(mlen>>8));
  UL_BLK_HEAD(blk).len=di.pos;

  blk=ul_new_frame_head(udrv, 0, 0, 0,
       UL_BFL_M2IN|UL_BFL_REC|
       UL_BFL_PRQ|UL_BFL_ARQ /*|UL_BFL_LNMM*/);
  if(blk==NULL) { ul_free_blk(udrv,beg_blk); return;};
  /* UL_BLK_HEAD(blk).len=mlen; */
  ul_tail_frame_head(beg_blk,blk);
  ul_bll_move_mes(&udrv->prep_bll,beg_blk);
};

void geninfoblk(ul_drv *udrv)
{
  static int seq_num=0;
  ul_mem_blk *blk, *beg_blk;
  ul_data_it di;
  blk=ul_new_frame_head(udrv, 3, 2, UL_CMD_SID,
       UL_BFL_M2IN|UL_BFL_TAIL|UL_BFL_SND|
       UL_BFL_PRQ|UL_BFL_ARQ);
  if(blk==NULL) return;
  beg_blk=blk;
  ul_di_init(&di,blk);
  /* ul_di_write1(&di,0); */
  ul_di_write1(&di,1);
  ul_di_write1(&di,2);
  ul_di_write1(&di,3);
  ul_di_write1(&di,(uchar)(seq_num++));
  UL_BLK_HEAD(blk).len=di.pos;

  blk=ul_new_frame_head(udrv, 0, 0, 0,
       UL_BFL_M2IN|UL_BFL_REC|
       UL_BFL_PRQ|UL_BFL_ARQ);
  if(blk==NULL) { ul_free_blk(udrv,beg_blk); return;};
  ul_tail_frame_head(beg_blk,blk);
  ul_bll_move_mes(&udrv->prep_bll,beg_blk);
};

void gentestblk(ul_drv *udrv)
{
  static int seq_num=0;
  ul_mem_blk *blk;
  ul_data_it di;
  blk=ul_new_frame_head(udrv, 3, 2, 0x23,
       UL_BFL_M2IN|UL_BFL_SND|UL_BFL_ARQ);
  if(blk==NULL) return;
  ul_di_init(&di,blk);
  /* ul_di_write1(&di,0); */
  ul_di_write1(&di,0x11);
  ul_di_write1(&di,0x12);
  ul_di_write1(&di,0x13);
  ul_di_write1(&di,(uchar)(seq_num++));
  UL_BLK_HEAD(blk).len=di.pos;

  ul_bll_move_mes(&udrv->prep_bll,blk);
};

#endif /* UL_INT_TST */
	
void printblk(ul_mem_blk *blk)
{
  ul_mem_blk *prev_blk=NULL;
  while(blk)
  {
    if(prev_blk) if(UL_BLK_HEAD(blk).prev!=prev_blk)
       UL_PRINTF(KERN_CRIT "Inconsistency in BLL list !!!\n");

    UL_PRINTF("BLK : dadr=%02X sadr=%02X cmd=%02X flg=%04X retry=%02X len=%04X\n",
     UL_BLK_HEAD(blk).dadr,
     UL_BLK_HEAD(blk).sadr,
     UL_BLK_HEAD(blk).cmd,
     UL_BLK_HEAD(blk).flg,
     UL_BLK_HEAD(blk).retry_cnt,
     UL_BLK_HEAD(blk).len
    );
    {
      ul_data_it di;
      uchar ch;
      int cnt=128;
      UL_PRINTF("      |");
      ul_di_init(&di,blk);
      di.trans_len=UL_BLK_HEAD(blk).len;
      while((di.pos<di.trans_len)&& --cnt)
      {
	ch=ul_di_read1(&di);
	if(ch>=0x20&&ch<0x7F) UL_PRINTF("%c",ch);
	else UL_PRINTF("<%X>",ch);
      };
      if(!cnt) UL_PRINTF("<...>...");
      UL_PRINTF("|\n");
    };
    prev_blk=blk;
    blk=UL_BLK_HEAD(blk).next;
  };
};

void printudrvbll(ul_drv *udrv)
{
  UL_PRINTF("\nprep_bll : frames = %d\n",udrv->prep_bll.cnt);
  printblk(udrv->prep_bll.first);

  UL_PRINTF("\nwork_bll : frames = %d\n",udrv->work_bll.cnt);
  printblk(udrv->work_bll.first);

  UL_PRINTF("\nproc_bll : frames = %d\n",udrv->proc_bll.cnt);
  printblk(udrv->proc_bll.first);
  UL_PRINTF("\nopan_bll : frames = %d\n",udrv->opan_bll.cnt);
  printblk(udrv->opan_bll.first);
  UL_PRINTF("\nfree_blk = %d",udrv->free_blk_cnt);
  UL_PRINTF("\n");
 #ifdef ENABLE_UL_MEM_CHECK
  UL_PRINTF("MEM_CHECK: malloc-free=%d\n",
		 (int)atomic_read(&ul_mem_check_counter));
 #endif /* ENABLE_UL_MEM_CHECK */
};

void printudrvoperators(ul_drv *udrv)
{
  UL_DRV_LOCK_FINI
  ul_opdata *opdata;
  ul_opchain *opmember,**opchain;
  int reccnt,filtcnt,opcnt;
  UL_PRINTF("\nul_drv operators (open handles)\n");
  UL_DRV_LOCK;
  opcnt=0;
  for(opdata=udrv->operators;opdata;opdata=opdata->opnext)
  { opchain=&opdata->recchain;reccnt=0;
    if((opmember=*opchain)!=NULL)do{
      reccnt++; if(reccnt>1000) break;
    }while((opmember=opmember->next)!=*opchain);
    opchain=&opdata->filtchain;filtcnt=0;
    if((opmember=*opchain)!=NULL)do{
      filtcnt++; if(filtcnt>1000) break;
    }while((opmember=opmember->next)!=*opchain);
    UL_PRINTF("  operator %d, in recchain %d, in filtchain %d\n",
		    opcnt,reccnt,filtcnt);
    opcnt++;if(opcnt>100) break;
  }
  UL_DRV_UNLOCK;
}

void printudrvfncstack(ul_drv *udrv)
{
  ul_call_fnc **p;
  UL_PRINTF("ul_drv: fnc_act:%08lx fnc_stack:",(long)udrv->fnc_act);
  p=udrv->fnc_sp;
  while(p!=&udrv->fnc_stack[0]){
    p--;
    UL_PRINTF("%08lx ",(long)*p);
  };
  UL_PRINTF("\n");
}

/*******************************************************************/
/* Linux kernel and module specific part */ 

#ifdef FOR_LINUX_KERNEL

MODULE_SUPPORTED_DEVICE("ul_drv");
MODULE_AUTHOR("Pavel Pisa <pisa@cmp.felk.cvut.cz>");
MODULE_DESCRIPTION("\
uLan communication module ( 9 bit RS-485 protocol )\n\
for communication with embedded controllers.\n\
It uses i82510 addon card or RS-232 dongle");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;

static int ulan_major_dev=ULAN_MAJOR+1;	/* uLan major device number */

#define UL_MINORS 9
ul_drv *ul_drv_arr[UL_MINORS]={[0 ... UL_MINORS-1]=NULL};
char *chip[UL_MINORS]={NULL,};
#if defined(UL_WITH_PCI) || defined(UL_WITH_USB)
char *slot[UL_MINORS]={NULL,};
#endif
int my_adr[UL_MINORS]={[0 ... UL_MINORS-1]=2};
int baud[UL_MINORS]={[0 ... UL_MINORS-1]=19200};
int irq[UL_MINORS]={0,};
int port[UL_MINORS]={0x3e8 
	#if (UL_MINORS>1)
		,[1 ... UL_MINORS-1]=0
	#endif
		};
int debug=-1;		/* debug flags */
int ul_usb_msg_inpr=0;	/* debug flags */

MODULE_PARM_DESC(chip,"chip type: auto, 82510, 16450, 16950-pci");
MODULE_PARM(chip,"1-" __MODULE_STRING(UL_MINORS) "s");
MODULE_PARM_DESC(my_adr,"uLan address of defined interface");
MODULE_PARM(my_adr,"1-" __MODULE_STRING(UL_MINORS) "i");
MODULE_PARM_DESC(baud,"define baudrates");
MODULE_PARM(baud,"1-" __MODULE_STRING(UL_MINORS) "i");
MODULE_PARM_DESC(irq,"define irqs");
MODULE_PARM(irq,"1-" __MODULE_STRING(UL_MINORS) "i");
MODULE_PARM_DESC(port,"define ports");
MODULE_PARM(port,"1-" __MODULE_STRING(UL_MINORS) "i");
#if defined(UL_WITH_PCI) || defined(UL_WITH_USB)
MODULE_PARM_DESC(slot,"slot type (pci) and specification");
MODULE_PARM(slot,"1-" __MODULE_STRING(UL_MINORS) "s");
#endif
MODULE_PARM_DESC(debug,"debug flags (FATAL,CHIO,IRQ,MSG,FAILS,"\
		       "SEQ,PORTS,FILE,FILT)");
MODULE_PARM(debug,"i");
MODULE_PARM_DESC(ul_usb_msg_inpr,"used length of Tx fifo in conventor");
MODULE_PARM(ul_usb_msg_inpr,"i");
MODULE_PARM_DESC(ulbuffer,"size of per interface buffer");
MODULE_PARM(ulbuffer,"i");

static int ulan_open(struct inode *inode, struct file *file);
static CLOSERET ulan_close(struct inode *inode, struct file *file);
static int ulan_ioctl(struct inode *inode, struct file *file, 
			unsigned int cmd, unsigned long arg);
static RWRET ulan_write(WRITE_PARAMETERS);
static RWRET ulan_read(READ_PARAMETERS);
static loff_t ulan_lseek(LSEEK_PARAMETERS);
static unsigned int ulan_poll(struct file *, poll_table *);

/* static struct file_operations ulan_fops */
KC_CHRDEV_FOPS_BEG(ulan_fops)
	KC_FOPS_LSEEK(ulan_lseek)/* lseek              */
	read:ulan_read,		/* read                */
	write:ulan_write,	/* write               */
	readdir:NULL,		/* readdir             */
	poll:ulan_poll,		/* poll                */
	ioctl:ulan_ioctl,	/* ioctl               */
	mmap:NULL,		/* mmap                */
	open:ulan_open,		/* open                */
	KC_FOPS_FLUSH(NULL)	/* flush               */
	KC_FOPS_RELEASE(ulan_close)/* close/release    */
	fsync:NULL,		/* fsync               */
	fasync:NULL,		/* fasync              */
	NULL,			/* check_media_change  */
	NULL,			/* revalidate          */
KC_CHRDEV_FOPS_END;

#ifdef ENABLE_UL_MEM_CHECK
void * ul_mem_check_malloc(size_t size)
{void *ptr;
 ptr=kmalloc(size,GFP_KERNEL);
 if(ptr) atomic_inc(&ul_mem_check_counter);
 return ptr;
}

void ul_mem_check_free(void *ptr)
{
  if(ptr) atomic_dec(&ul_mem_check_counter);
  kfree(ptr);
}
#endif /* ENABLE_UL_MEM_CHECK */
#endif /* FOR_LINUX_KERNEL */

static void add_to_opchain(ul_opchain **chain,ul_opchain *member)
{
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  if(!*chain) *chain=member->next=member->prev=member;
  else {
    (volatile ul_opchain*)member->next=*chain;
    (volatile ul_opchain*)member->prev=(*chain)->prev;
    UL_MB();
    (volatile ul_opchain*)(*chain)->prev->next=member;
    (volatile ul_opchain*)(*chain)->prev=member;
  };
  UL_DRV_UNLOCK;
};

static void add_to_opchain_front(ul_opchain **chain,ul_opchain *member)
{
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  if(!*chain) *chain=member->next=member->prev=member;
  else {
    (volatile ul_opchain*)member->next=*chain;
    (volatile ul_opchain*)member->prev=(*chain)->prev;
    UL_MB();
    (volatile ul_opchain*)(*chain)->prev->next=member;
    (volatile ul_opchain*)(*chain)->prev=member;
    *(volatile ul_opchain**)chain=member;
  };
  UL_DRV_UNLOCK;
};

static void del_from_opchain(ul_opchain **chain,ul_opchain *member)
{
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  member->next->prev=member->prev;
  member->prev->next=member->next;
  if(*chain==member) 
  { if(member->next==member) *chain=NULL;
    else *chain=member->next;
  };
  UL_DRV_UNLOCK;
};

/* free message form ulan operator */
static int ulan_freemsg(ul_opdata* opdata)
{
  int stamp;
  ul_mem_blk *mes;
  ul_opchain *opmember;

  if(!opdata->udrv) return -ENODEV;
  if(!(mes=opdata->message)) return -ENOMSG;

  if(!UL_BLK_HEAD(mes).stamp)
  {
    opdata->message=NULL;
    stamp=ul_gen_stamp();
    UL_BLK_HEAD(mes).stamp=stamp;
    if(UL_BLK_HEAD(mes).flg&UL_BFL_M2IN)
    {
      opmember=(ul_opchain*)MALLOC(sizeof(ul_opchain));
      if(opmember)
      {
	memset(opmember,0,sizeof(ul_opchain));
	opmember->stamp=stamp;
	opmember->state=UL_OPST_ONCE;
	add_to_opchain_front(&opdata->filtchain,opmember);
      };
    };
    ul_bll_move_mes(&opdata->udrv->prep_bll,mes);
    ulan_stroke(opdata->udrv);
    return stamp;
  }else{
    ul_dec_ref_cnt(mes);
    opdata->message=NULL;
    return 0;
  };
};

/* abort preparation of message */
static int ulan_abortmsg(ul_opdata* opdata)
{
  ul_mem_blk *mes;

  if(!opdata->udrv) return -ENODEV;
  if(!(mes=opdata->message)) return -ENOMSG;
  opdata->message=NULL;

  if(!UL_BLK_HEAD(mes).stamp)
  {
    ul_free_mes(opdata->udrv,mes);
    return 1;
  }else{
    ul_dec_ref_cnt(mes);
    return 0;
  };
};

/* rewind of message */
static int ulan_rewmsg(ul_opdata* opdata)
{
  ul_mem_blk *mes;
  if(!opdata->udrv) return -ENODEV;
  if(!(mes=opdata->message)) return -ENOMSG;
  opdata->message=NULL;
  ul_di_init(&opdata->data,opdata->message);
  return 0;
};

/* create new outgoing message */
static int ulan_newmsg(ul_opdata* opdata,const ul_msginfo *msginfo)
{
  int flg=msginfo->flg;
  flg &= ~(UL_BFL_LOCK | UL_BFL_FAIL | UL_BFL_TAIL | UL_BFL_REC);
  flg |= UL_BFL_SND;
  if(!opdata->udrv) return -ENODEV;
  if (opdata->message) ulan_freemsg(opdata);
  opdata->message=ul_new_frame_head(opdata->udrv, msginfo->dadr,
                     opdata->udrv->my_adr, msginfo->cmd,flg);
  if(!opdata->message) return -ENOMEM;
  memset(UL_BLK_FDATA(opdata->message),0,UL_BLK_FSIZE);
  ul_di_init(&opdata->data,opdata->message);
  return 0;
}

static int ulan_tailmsg(ul_opdata* opdata,const ul_msginfo *msginfo)
{
  ul_mem_blk *mes;
  ul_mem_blk *tail;
  int flg=msginfo->flg;
  if(!opdata->udrv) return -ENODEV;
  if(!(mes=opdata->message)) return -ENOMSG;
  if(UL_BLK_HEAD(mes).stamp) return -EPERM;

  flg &= ~(UL_BFL_LOCK | UL_BFL_FAIL | UL_BFL_TAIL);
  if(!flg&UL_BFL_REC) flg|=UL_BFL_SND;
  tail=ul_new_frame_head(opdata->udrv, msginfo->dadr,
	opdata->udrv->my_adr, msginfo->cmd,flg);
  if(!tail) return -ENOMEM;
  memset(UL_BLK_FDATA(tail),0,UL_BLK_FSIZE);
  UL_BLK_HEAD(tail).len=msginfo->len;
  ul_tail_frame_head(mes,tail);
  ul_di_init(&opdata->data,tail);
  return 0;
}

static int ulan_acceptmsg(ul_opdata* opdata,ul_msginfo *msginfo)
{
  ul_mem_blk *mes;
  ul_opchain *opmember;

  if(!opdata->udrv) return -ENODEV;
  if (opdata->message) ulan_freemsg(opdata);
  memset(msginfo,0,sizeof(ul_msginfo));
  if(!(opmember=opdata->recchain)) return -ENOMSG;
  del_from_opchain(&opdata->recchain,opmember);
  mes=opmember->message;
  if(opmember->state==UL_OPST_MES) FREE(opmember);
  opdata->message=mes;
  ul_di_init(&opdata->data,mes);
  msginfo->flg=UL_BLK_HEAD(mes).flg;
  msginfo->dadr=UL_BLK_HEAD(mes).dadr;
  msginfo->sadr=UL_BLK_HEAD(mes).sadr;
  msginfo->cmd=UL_BLK_HEAD(mes).cmd;
  msginfo->len=UL_BLK_HEAD(mes).len;
  msginfo->stamp=UL_BLK_HEAD(mes).stamp;
  return 0;
}

static int ulan_actailmsg(ul_opdata* opdata,ul_msginfo *msginfo)
{
  ul_mem_blk *mes;
  ul_mem_blk *tail;

  if(!opdata->udrv) return -ENODEV;
  memset(msginfo,0,sizeof(ul_msginfo));
  if(!(mes=opdata->message)) return -ENOMSG;
  if(!UL_BLK_HEAD(mes).stamp) return -EPERM;
  tail=UL_BLK_HEAD(opdata->data.head_blk).next;
  if(!tail) return -ENOMSG;
  ul_di_init(&opdata->data,tail);
  msginfo->flg=UL_BLK_HEAD(tail).flg;
  msginfo->dadr=UL_BLK_HEAD(tail).dadr;
  msginfo->sadr=UL_BLK_HEAD(tail).sadr;
  msginfo->cmd=UL_BLK_HEAD(tail).cmd;
  msginfo->len=UL_BLK_HEAD(tail).len;
  msginfo->stamp=UL_BLK_HEAD(tail).stamp;
  return 0;
}

/* create new outgoing message */
static int ulan_addfilt(ul_opdata* opdata,const ul_msginfo *msginfo)
{
  ul_drv *udrv;
  ul_opchain *opmember;
  int filtonce;
  if(!opdata->udrv) return -ENODEV;
  opmember=(ul_opchain*)MALLOC(sizeof(ul_opchain));
  udrv=opdata->udrv;
  if(!opmember) return -ENOMEM;
  memset(opmember,0,sizeof(ul_opchain));
  opmember->dadr=msginfo->dadr;
  opmember->sadr=msginfo->sadr;
  opmember->cmd=msginfo->cmd;
  filtonce=msginfo->flg&UL_BFL_NORE;
  opmember->state=filtonce?UL_OPST_FILTNEW:UL_OPST_FILT;
  add_to_opchain(&opdata->filtchain,opmember);
  if(filtonce){
    uld_set_dfl(udrv,CHECK_FILT);
    SCHEDULE_BH(udrv);
  }
  LOG_FILTS("ulan_addfilt : filter %d %d %d %s\n",
	 opmember->dadr,opmember->sadr,opmember->cmd,filtonce?"once":"");
  return 0;
}

static int ulan_inepoll(ul_opdata* opdata)
{
  return opdata->recchain!=NULL?1:0;
}

static int ulan_hwtest(ul_opdata* opdata,unsigned long fnc)
{
  int ret;
  ul_drv *udrv;
  UL_DRV_LOCK_FINI
  if(!opdata->udrv) return -ENODEV;
  udrv=opdata->udrv;
  UL_DRV_LOCK;
  ret=udrv->fnc_cctrl(udrv,UL_CC_HWTEST,fnc);
  if(ret<0) ret=-EINVAL;
  UL_DRV_UNLOCK;
  return ret;
}

#ifdef FOR_LINUX_KERNEL

/* open ulan device */
static int ulan_open(struct inode *inode, struct file *file)
{ 
  ul_drv *udrv;
  ul_opdata *opdata;
  unsigned int minor = kc_dev2minor(inode->i_rdev);
  #ifdef UL_WITH_DEVFS
  if(minor==-1){
    udrv=file->private_data;
  }else
  #endif /* UL_WITH_DEVFS */
  {
    if ((minor >= UL_MINORS)||(minor<0)) return -ENODEV;
    udrv=ul_drv_arr[minor];
  }
  if (!udrv) return -ENODEV;
  if(udrv->magic!=UL_DRV_MAGIC)
  {
    printk(KERN_CRIT "ulan_open: Wrong uLan MAGIC number!!!\n");
    return -ENODEV;
  }

  kc_MOD_INC_USE_COUNT;
  if(!(file->private_data=MALLOC(sizeof(ul_opdata))))
  { kc_MOD_DEC_USE_COUNT;
    return(-ENOMEM);
  };
  opdata=(ul_opdata*)file->private_data;
  opdata->file=file;
  opdata->magic=ULOP_MAGIC;
  opdata->message=NULL;
  opdata->udrv=udrv;
  init_waitqueue_head(&opdata->wqrec);
  opdata->opprew=NULL;
  opdata->opnext=NULL;
  opdata->recchain=NULL;
  opdata->filtchain=NULL;
  { /* add us onto list of clients of udrv */
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    opptr=udrv->operators;
    if(opptr) {opptr->opprew=opdata;opdata->opnext=opptr;};
    UL_MB();
    udrv->operators=opdata;
    UL_DRV_UNLOCK;
  };

  return 0;
};

static CLOSERET ulan_close(struct inode *inode, struct file *file)
{
  ul_opchain *opmember;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_close : BAD opdata magic !!!");

  if (opdata->message) ulan_freemsg(opdata);
  { /* delete us from list of clients of udrv */
    ul_drv *udrv;
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    if((udrv=opdata->udrv)!=NULL){
      if((opptr=opdata->opnext)) opptr->opprew=opdata->opprew;
      if((opptr=opdata->opprew)) opptr->opnext=opdata->opnext;
      else udrv->operators=opdata->opnext;
    }
    UL_DRV_UNLOCK;
    if(udrv)
      if(udrv->irq>0) ul_synchronize_irq(udrv->irq);
    while((udrv=opdata->udrv)!=NULL){
      if(!uld_atomic_test_dfl(udrv,IN_BOTTOM))
        break;
      kc_yield();
    }
  }
  schedule();
  while((opmember=opdata->recchain))
  {
    del_from_opchain(&opdata->recchain,opmember);
    if(opmember->message) ul_dec_ref_cnt(opmember->message);
    FREE(opmember);
  };
  while((opmember=opdata->filtchain))
  {
    del_from_opchain(&opdata->filtchain,opmember);
    FREE(opmember);
  };
  FREE(file->private_data);
  kc_MOD_DEC_USE_COUNT;
  return (CLOSERET)0;
};

static int ulan_ioctl(struct inode *inode, struct file *file, 
			unsigned int cmd, unsigned long arg)
{
  ul_msginfo msginfo;
  int ret;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_ioctl : BAD opdata magic !!!");

  switch (cmd) {
   case UL_DRV_VER :
	return UL_DRV_VERCODE;
   case UL_NEWMSG :
	if(kc_copy_from_user(&msginfo,(void *)arg,sizeof(msginfo)))
	   return -EFAULT;
  	return ulan_newmsg(opdata,&msginfo);
   case UL_TAILMSG :
	if(kc_copy_from_user(&msginfo,(void *)arg,sizeof(msginfo)))
	   return -EFAULT;
  	return ulan_tailmsg(opdata,&msginfo);
   case UL_FREEMSG :
  	return ulan_freemsg(opdata);
   case UL_ACCEPTMSG :  	
	if(kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo)))
	   return -EFAULT;
	ret=ulan_acceptmsg(opdata,&msginfo);
	kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo));
  	return  ret;
   case UL_ACTAILMSG :  	
	if(kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo)))
	   return -EFAULT;
	ret=ulan_actailmsg(opdata,&msginfo);
	kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo));
  	return  ret;
   case UL_ADDFILT :
	if(kc_copy_from_user(&msginfo,(void *)arg,sizeof(msginfo)))
	   return -EFAULT;
  	return ulan_addfilt(opdata,&msginfo);
   case UL_ABORTMSG :
  	return ulan_abortmsg(opdata);
   case UL_REWMSG :
  	return ulan_rewmsg(opdata);
   case UL_KLOGBLL :
        if(!opdata->udrv) return -ENODEV;
	printudrvbll(opdata->udrv);
    #ifdef ENABLE_UL_MEM_CHECK
	printudrvoperators(opdata->udrv);
    #endif /* ENABLE_UL_MEM_CHECK */
	return 0;
   case UL_STROKE :
        if(!opdata->udrv) return -ENODEV;
	ulan_stroke(opdata->udrv);
	return 0;
   case UL_DEBFLG :
	uld_debug_flg=arg;
	return 0;
   case UL_INEPOLL :
        return ulan_inepoll(opdata);
   case UL_HWTEST :
        if(current->euid!=0) return -EPERM;
	return ulan_hwtest(opdata,arg);
  };


  return -EINVAL;
};

/* common parameters (struct file *file, const char *buf, 
                      size_t count [, loff_t *ppos]) */
static RWRET ulan_write(WRITE_PARAMETERS)
{
  int cn;
  int len;
  ul_mem_blk *blk;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_write : BAD opdata magic !!!");

  if(!opdata->udrv) return -ENODEV;
  if(!opdata->message) return -ENOMSG;

  cn=count;
  while(cn>0)
  {
    while(!ul_di_adjust(&opdata->data))
    {
      if(!(blk=ul_alloc_blk(opdata->udrv))) 
	{count-=cn;cn=0;break;};
      memset(UL_BLK_NDATA(blk),0,UL_BLK_SIZE);
      opdata->data.blk->next=blk;
    };
    len=ul_di_atonce(&opdata->data);
    if(len>cn) len=cn;
    kc_copy_from_user(ul_di_byte(&opdata->data),buf,len);
    ul_di_add(&opdata->data,len);
    buf+=len;
    cn-=len;
  };

  if(opdata->data.pos>UL_BLK_HEAD(opdata->data.head_blk).len)
    UL_BLK_HEAD(opdata->data.head_blk).len=opdata->data.pos;

  return count-cn;
};

/* common parameters (struct file *file, char *buf, 
                      size_t count [, loff_t *ppos]) */
static RWRET ulan_read(READ_PARAMETERS)
{
  int cn;
  int len;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_read : BAD opdata magic !!!");

  if(!opdata->udrv) return -ENODEV;
  if(!opdata->message) return -ENOMSG;

  if(opdata->data.pos+count>UL_BLK_HEAD(opdata->data.head_blk).len)
   count=UL_BLK_HEAD(opdata->data.head_blk).len-opdata->data.pos;

  cn=count;
  while(cn>0)
  {
    if(!ul_di_adjust(&opdata->data))
    {
      while(cn--) kc_put_user(0,buf++);
      break;
    };
    len=ul_di_atonce(&opdata->data);
    if(len>cn) len=cn;
    kc_copy_to_user(buf,ul_di_byte(&opdata->data),len);
    ul_di_add(&opdata->data,len);
    buf+=len;
    cn-=len;
  };

  return count-cn;
};

/* common parameters (struct file *file, loff_t pos, int whence) */
static loff_t ulan_lseek(LSEEK_PARAMETERS)
{
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_lseek : BAD opdata magic !!!");

  if(!opdata->message) return -ENOMSG;

  return ul_di_seek(&opdata->data,pos,whence);
};

static unsigned int ulan_poll(struct file *file, poll_table *wait)
{
  /* POLLIN | POLLOUT | POLLERR | POLLHUP | POLLRDNORM | POLLWRNORM */
  unsigned int mask = 0;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_poll : BAD opdata magic !!!");

  kc_poll_wait(file, &opdata->wqrec, wait);

  if (opdata->recchain)
    mask |= (POLLIN | POLLRDNORM);

  if (!opdata->udrv)
    mask |= (POLLIN | POLLOUT | POLLERR);

  return mask;
};

#endif /* FOR_LINUX_KERNEL */


/* this function is responsible to inform all ulan client operators
   about message of waited type -- !!!!! it is not SMP clean still */
static int ulan_proc_arrived(ul_drv *udrv, ul_mem_blk *message)
{
  ul_opdata *opdata;
  ul_opchain **opchain;
  ul_opchain *opmember;
  ul_opchain *opmember2;
  UL_DRV_LOCK_FINI

  ul_inc_ref_cnt(message);
  UL_DRV_LOCK;
  opdata=udrv->operators;
  UL_DRV_UNLOCK;
  while(opdata)
  {
    opchain=&opdata->filtchain;
    opmember=*opchain;
    if(opmember) do
    {
      if((!opmember->stamp||opmember->stamp==UL_BLK_HEAD(message).stamp)&&
	 (!opmember->cmd||opmember->cmd==UL_BLK_HEAD(message).cmd)&&
	 (!opmember->dadr||opmember->dadr==UL_BLK_HEAD(message).dadr)&&
	 (!opmember->sadr||opmember->sadr==UL_BLK_HEAD(message).sadr))
      {
	if(opmember->state==UL_OPST_ONCE)
	{
	  del_from_opchain(opchain,opmember);
	  opmember->state=UL_OPST_MES;
	  ul_inc_ref_cnt(message);
	  opmember->message=message;
	  add_to_opchain(&opdata->recchain,opmember);
	}else if(opmember->state==UL_OPST_FILT){
	 #ifdef FOR_LINUX_KERNEL
	  #ifdef ENABLE_UL_MEM_CHECK
	  atomic_inc(&ul_mem_check_counter);
	  #endif
	  opmember2=(ul_opchain*)kmalloc(sizeof(ul_opchain),GFP_ATOMIC);
	 #else /* FOR_LINUX_KERNEL */
	  opmember2=(ul_opchain*)MALLOC(sizeof(ul_opchain));
	 #endif /* FOR_LINUX_KERNEL */
	  if(opmember2)
	  { 
	    memset(opmember2,0,sizeof(ul_opchain));
	    opmember2->state=UL_OPST_MES;
	    ul_inc_ref_cnt(message);
	    opmember2->message=message;
	    add_to_opchain(&opdata->recchain,opmember2);
	  } else  UL_PRINTF(KERN_CRIT "ulan_proc_arrived : no kernel memory : FIXME handling\n");
	};
       #ifdef FOR_LINUX_KERNEL
	wake_up_interruptible(&opdata->wqrec);
       #elif defined(_WIN32)
	    wake_and_xch_pending_irp(&opdata->wait_irp,NULL,STATUS_SUCCESS);
       #endif /* FOR_LINUX_KERNEL */
	break;
      };
    } while((opmember=opmember->next)!=*opchain);
    opdata=opdata->opnext;
  };

  /* if no one interested in message, it is discarded */
  return  ul_dec_ref_cnt(message);
};

/* check for UL_OPST_FILTNEW and change them to single instance
   of UL_OPST_FILT */
void check_for_filtnew(ul_drv *udrv)
{
  ul_opdata *opdata;
  ul_opchain **opchain;
  ul_opchain *opmember;
  ul_opchain *opmember2;
  ul_opchain *opnext;
  UL_DRV_LOCK_FINI

  LOG_FILTS("check_for_filtnew : called\n");
  UL_DRV_LOCK;
  opdata=udrv->operators;
  UL_DRV_UNLOCK;
  for(;opdata;opdata=opdata->opnext)
  {
    opchain=&opdata->filtchain;
    opmember=*opchain;
    if(opmember) do
    { opnext=opmember->next;
      if(opnext==*opchain) opnext=NULL;
      if(opmember->state==UL_OPST_FILTNEW){
	LOG_FILTS("check_for_filtnew : processing UL_OPST_FILTNEW\n");
	opmember2=*opchain;
	while(opmember2){
	  if((opmember->cmd==opmember2->cmd)&&
             (opmember->sadr==opmember2->sadr)&&
	     (opmember->dadr==opmember2->dadr)&&
	     (opmember->stamp==opmember2->stamp)&&
	     (opmember2->state==UL_OPST_FILT)) break;
	  opmember2=opmember2->next;
	  if(opmember2==*opchain){
	    opmember2=NULL;
	  }
        };
	if(opmember2){
	  LOG_FILTS("check_for_filtnew : deletting UL_OPST_FILTNEW\n");
	  del_from_opchain(opchain,opmember);
	  FREE(opmember);
	}else{
	  LOG_FILTS("check_for_filtnew : changing to UL_OPST_FILT\n");
	  opmember->state=UL_OPST_FILT;
	}
      }
    } while((opmember=opnext)!=NULL);
  };
};

#ifndef _WIN32

/* notice about new message in udrv->proc_bll */
void ulan_do_bh(ulan_do_bh_data_type data)
{
  ul_mem_blk *message;
  ul_drv *udrv=(ul_drv *)data;

  #ifdef ENABLE_UL_IRQ_STALE_WDG
  ul_irq_stale_wdg_cnt=0;
  #endif /* ENABLE_UL_IRQ_STALE_WDG */

  if(udrv->magic!=UL_DRV_MAGIC)
  {
    #ifdef FOR_LINUX_KERNEL
     panic("ulan_do_bh : BAD udrv magic !!!");
    #else /* FOR_LINUX_KERNEL */
     return;
    #endif /* FOR_LINUX_KERNEL */
  }

  LOG_MESSAGES("ulan_do_bh : announcing messages\n");
  while(udrv->proc_bll.first||uld_test_dfl(udrv,ASK_BOTTOM))
  {
    if(uld_test_and_set_dfl(udrv,IN_BOTTOM))
    { LOG_MESSAGES("ulan_do_bh : else is already in");
	  return;
    }
    uld_clear_dfl(udrv,ASK_BOTTOM);
    if(uld_test_dfl(udrv,CHECK_FILT)){
      uld_clear_dfl(udrv,CHECK_FILT);
      check_for_filtnew(udrv);
    }
    while((message=udrv->proc_bll.first))
    {
      int i;
      ul_bll_move_mes(&udrv->opan_bll,message);
      i=ulan_proc_arrived(udrv,message);
      LOG_MESSAGES("ulan_do_bh : message sent to %d recchains\n",i);
    };
    uld_atomic_clear_dfl(udrv,IN_BOTTOM);
  };
};

#endif /* _WIN32 */

#ifdef FOR_LINUX_KERNEL

/* config space access (byte|word|dword) return to text pcibios_strerror */
/* pci_read_config_byte(struct pci_dev *dev, int where, u8 *val) */
/* pci_write_config_byte(struct pci_dev *dev, int where, u8 *val) */
/* request_region(), request_mem_region() */
/* void *pci_get_drvdata (struct pci_dev *pdev) */
/* void pci_set_drvdata (struct pci_dev *pdev, void *data) */


#if defined(UL_WITH_PCI) || defined(UL_WITH_USB)

static int /*__devinit*/ 
 ulan_init_find_minor(char *abus, char *aslot, char *asubdev,
                      int *pminor, int *ppar)
{
  int i;
  int abus_len;
  int aslot_len;
  int sub_offs;
  int minor;
  int par;
  int match=0;
  char *islot;
  abus_len=strlen(abus);
  aslot_len=strlen(aslot);
  sub_offs=abus_len+aslot_len;
  *pminor=*ppar=minor=par=-1;
  for(i=0;i<UL_MINORS;i++){
    islot=slot[i];
    if(islot)
      if(!*islot) islot=NULL;
    switch (match){
      case 0: 
        if(!ul_drv_arr[i]&&!islot){
	  minor=i;match=1;
	  if(par==-1) par=i;
	}
      case 1:
        if(!islot) continue;
        if(!strcmp(islot,abus)){
	  par=i;
	  if(!ul_drv_arr[i]){ minor=i; match=2;}
	}
      case 2:
        if(!islot||!aslot) continue;
        if(!strncmp(islot,abus,abus_len)&&
	   !strcmp(islot+abus_len,aslot)){
	  par=i;
	  if(!ul_drv_arr[i]){ minor=i; match=3;}
        }
      case 3:
        if(!islot||!aslot||!asubdev) continue;
        if(!strncmp(islot,abus,abus_len)&&
	   !strncmp(islot+abus_len,aslot,aslot_len)&&
	   (islot[sub_offs]=='.')&&!strcmp(islot+sub_offs+1,asubdev)){
	  par=i;
	  if(!ul_drv_arr[i]){ minor=i; match=4;}
        }
    }
  }
  *pminor=minor;*ppar=par;
  return match;
}

#endif /*defined(UL_WITH_PCI)||defined(UL_WITH_USB)*/

/*******************************************************************/
/* PCI Devices Support */

#ifdef UL_WITH_PCI

/* default invocation of chip_init for PCI devices */
static int /*__devinit*/ 
  ulan_init_chan(struct pci_dev *dev,char *subdev, 
  		 char *chip_name, ul_drv **pudrv,
                 ul_chip_init_fnc *chip_init,
                 int port, int irq, int options)
{
  ul_drv *udrv;
  int amy_adr=0;
  int abaud=0;
  int match;
  int minor;
  int ret;
  int i;
#ifdef UL_WITH_DEVFS
  kc_devfs_handle_t devfs_handle;
  char dev_name[32];
#endif /* UL_WITH_DEVFS */
  
  *pudrv=NULL;
  /* try to find best minor and parameters */
  match=ulan_init_find_minor("pci",dev->slot_name,subdev,&minor,&i);
  if(i>=0){
    abaud=baud[i];amy_adr=my_adr[i];
  }

  /* mem for ul_drv */
  if(!(udrv=MALLOC(sizeof(ul_drv)))) return -ENOMEM;
  /* clear memory */
  memset(udrv,0,sizeof(ul_drv));
  /* set initial state */
  ul_drv_new_init_state(udrv, amy_adr);
  udrv->dev=dev;
  /* init chip driver */
  if((ret=(*chip_init)(udrv, port, irq, abaud, options))<0){
    printk(KERN_CRIT "ulan_init_chan: ERROR - chip_init returned %d\n",ret);
    FREE(udrv);
    return -EIO;
  }
  /* setups buffers, ports and irq for sucesfully detected device */
  if((ret=ul_drv_new_start(udrv,ulbuffer))<0){
    printk(KERN_CRIT "ulan_init_chan: ERROR - ul_drv_new_start returned %d\n",ret);
    FREE(udrv);
    return -EIO;
  }
  #ifdef UL_WITH_DEVFS
  sprintf (dev_name, "ulan%d", minor);
  devfs_handle=kc_devfs_new_cdev(NULL, MKDEV(ulan_major_dev, minor), 
			S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, 
			&ulan_fops, udrv, dev_name);
  udrv->devfs_handle=devfs_handle;
  #endif /* UL_WITH_DEVFS */

  printk(KERN_INFO "ulan_init_chan: chip=%s minor=%d baud=%d my_adr=%d ready\n",
	 chip_name,/*dev_name*/minor,udrv->baud_val,udrv->my_adr);
  
  if(minor>=0) ul_drv_arr[minor]=udrv;
  *pudrv=udrv;
  udrv->next_chan=pci_get_drvdata(dev);
  pci_set_drvdata(dev, udrv);
  return 0;
}


static int /*__devinit*/ ulan_init_one_950pci (struct pci_dev *dev,
				   const struct pci_device_id *ent)
{
  int chip_options=ent->driver_data;
  int ret, retall=-ENODEV;
  char subdev[2];
  int subno, port, irq;
  ul_drv *udrv;

  if (pci_enable_device (dev)){
    printk(KERN_CRIT "ulan_init_one_950pci: Cannot enable device\n");
    return -EIO;    
  }
  port=pci_resource_start(dev,0);
  irq=dev->irq;
  if(!irq||!port||!(pci_resource_flags(dev,0)&IORESOURCE_IO)){
    LOG_FATAL(KERN_CRIT "uLan u950pci_init : bad PCI port or irq !\n");
    return -ENODEV;
  }
  
  for(subno=0;subno<2;subno++,port+=8){
    subdev[0]='0'+subno;subdev[1]=0;
    /* (dev,subdev,chip_name,pudrv,chip_init,port,irq,options) */
    ret=ulan_init_chan(dev,subdev,"16950-pci",&udrv,&u950pci_init,port,irq,chip_options);
    if(ret>=0){
      retall=0;
    } else {
      if(retall<0) retall=ret;
    }
  }
  return retall;
}

static int /*__devinit*/ ulan_init_one (struct pci_dev *dev,
				   const struct pci_device_id *ent)
{
  unsigned long driver_data=ent->driver_data;
  printk(KERN_INFO "ulan_init_one: PCI device found at slot : %s\n",
         dev->slot_name?dev->slot_name:"unknown");
  pci_set_drvdata(dev, NULL);
  switch(driver_data&~0xff){
    case 0x16954000 :
      return ulan_init_one_950pci(dev,ent);
  }
  printk(KERN_CRIT "ulan_init_one: No device of specified driver_data type\n");
  return -ENODEV;
}

static void /*__devexit*/ ulan_remove_one (struct pci_dev *dev)
{
  ul_drv *udrv, *next_udrv;
  int i;
  udrv=(ul_drv *)pci_get_drvdata(dev);
  if(!udrv){
    printk(KERN_CRIT "ulan_remove_one: no uLan drvdata\n");
    return;
  }
  for(;udrv;udrv=next_udrv){
    if(udrv->magic!=UL_DRV_MAGIC){
      printk(KERN_CRIT "ulan_remove_one: Wrong uLan MAGIC number!!!\n");
      return;
    }
    next_udrv=udrv->next_chan;
    if(dev!=udrv->dev){
      printk(KERN_CRIT "ulan_remove_one: BAD - cross PCI device remove\n");
    }
    for(i=0;i<UL_MINORS;i++){
      if (udrv==ul_drv_arr[i]) ul_drv_arr[i]=NULL;
    }
    #ifdef UL_WITH_DEVFS
    if(udrv->devfs_handle) kc_devfs_delete(udrv->devfs_handle);
    #endif /* UL_WITH_DEVFS */
    ul_drv_free(udrv);
  }
  pci_set_drvdata(dev, NULL);
  printk(KERN_INFO "ulan_remove_one: PCI device removed\n");
}

static struct pci_device_id ulan_pci_tbl[] /*__devinitdata*/ = {
	/* Advantech PCI-1602A, negated DTR */
	{ /* PCI_VENDOR_ID */ 0x13fe, /* PCI_DEVICE_ID */ 0x1600,
	  /* PCI_VENDOR_ID */ PCI_ANY_ID, /* PCI_SUBDEVICE_ID */ PCI_ANY_ID,
	  0, 0, 0x16954000|U950PCI_CHOPT_TXDTRNEG|U950PCI_CHOPT_HSPDOSC},
	/* Tedia PCI-1482 */
	{ /* PCI_VENDOR_ID */ 0x1415, /* PCI_DEVICE_ID */ 0x950A,
	  /* PCI_VENDOR_ID */ PCI_ANY_ID, /* PCI_SUBDEVICE_ID */ PCI_ANY_ID,
	  0, 0, 0x16954000|U950PCI_CHOPT_RXDONRI},
	/* Tedia PCI-1482 with PiKRON signature */
	{ /* PCI_VENDOR_ID */ 0x1760, /* PCI_DEVICE_ID */ 0x8004,
	  /* PCI_VENDOR_ID */ PCI_ANY_ID, /* PCI_SUBDEVICE_ID */ PCI_ANY_ID,
	  0, 0, 0x16954000|U950PCI_CHOPT_RXDONRI},
	{ 0, } /* terminate list */
};

MODULE_DEVICE_TABLE(pci, ulan_pci_tbl);

static struct pci_driver ulan_pci_driver = {
	name:		"ul_drv",
	id_table:	ulan_pci_tbl,
	probe:		ulan_init_one,
	remove:		ulan_remove_one,
};

#endif  /* UL_WITH_PCI */


/*******************************************************************/
/* USB Devices Support */

#ifdef UL_WITH_USB

/* Define these values to match your device */
#define UL_USB_VENDOR_ID	0x4343
#define UL_USB_PRODUCT_ID	0x0001
#define UL_USB_HW_PS1		(0x1234<<8)

#if (LINUX_VERSION_CODE < VERSION(2,5,41))
static void *ul_usb_probe(struct usb_device *dev, unsigned int ifnum, const struct usb_device_id *id);
static void ul_usb_disconnect(struct usb_device *dev, void *ptr);
#else /* 2.5.41 */
static int ul_usb_probe(struct usb_interface *intf, const struct usb_device_id *id);
static void ul_usb_disconnect(struct usb_interface *intf);
#endif /* 2.5.41 */

int ul_usb_ps1_init(ul_drv *udrv, int port, int irq, int baud, int options);

/* table of devices that work with this driver */
static struct usb_device_id ul_usb_table [] = {
	{ USB_DEVICE(UL_USB_VENDOR_ID, UL_USB_PRODUCT_ID), driver_info:UL_USB_HW_PS1 },
	{ }					/* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, ul_usb_table);


#ifdef CONFIG_USB_DYNAMIC_MINORS
/* 
 * if the user wants to use dynamic minor numbers, then we can have up to 256
 * devices
 */
#define UL_USB_MINOR_BASE	0
#define MAX_DEVICES		256
#else
/* Get a minor range for your devices from the usb maintainer */
#define UL_USB_MINOR_BASE	200

/* we can have up to this number of device plugged in at once */
#define MAX_DEVICES		16
#endif

/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver ul_usb_driver = {
	name:		"ulan_usb",
	probe:		ul_usb_probe,
	disconnect:	ul_usb_disconnect,
	/*fops:		&ul_usb_fops,*/
	/* num_minors:	MAX_DEVICES,*/
     #if (LINUX_VERSION_CODE < VERSION(2,5,8))
	minor:		UL_USB_MINOR_BASE,
     #endif
	id_table:	ul_usb_table,
};

#endif /* UL_WITH_USB */



static int /*__devinit*/ 
  ul_usb_init_chan(struct usb_device *dev, unsigned int ifnum,
                 const struct usb_device_id *ent, ul_drv **pudrv,
                 ul_chip_init_fnc *chip_init, int options)
{
  ul_drv *udrv;
  int amy_adr=0;
  int abaud=0;
  int match;
  int minor;
  int ret;
  int i;
#ifdef UL_WITH_DEVFS
  kc_devfs_handle_t devfs_handle;
  char dev_name[32];
#endif /* UL_WITH_DEVFS */
  
  *pudrv=NULL;
  /* try to find best minor and parameters */
  match=ulan_init_find_minor("usb","unknown","unknown",&minor,&i);
  if(i>=0){
    abaud=baud[i];amy_adr=my_adr[i];
  }

  /* mem for ul_drv */
  if(!(udrv=MALLOC(sizeof(ul_drv)))) return -ENOMEM;
  /* clear memory */
  memset(udrv,0,sizeof(ul_drv));
  /* set initial state */
  ul_drv_new_init_state(udrv, amy_adr);
  udrv->dev=(struct pci_dev *)dev;
  /* init chip driver */
  if((ret=(*chip_init)(udrv, ifnum, 0, abaud, options))<0){
    printk(KERN_CRIT "ul_usb_init_chan: ERROR - chip_init returned %d\n",ret);
    FREE(udrv);
    return -EIO;
  }
  /* setups buffers, ports and irq for sucesfully detected device */
  if((ret=ul_drv_new_start(udrv,ulbuffer))<0){
    printk(KERN_CRIT "ulan_init_chan: ERROR - ul_drv_new_start returned %d\n",ret);
    FREE(udrv);
    return -EIO;
  }
  #ifdef UL_WITH_DEVFS
  sprintf (dev_name, "ul_usb%d", minor);
  devfs_handle=kc_devfs_new_cdev(NULL, MKDEV(ulan_major_dev, minor), 
			S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, 
			&ulan_fops, udrv, dev_name);
  udrv->devfs_handle=devfs_handle;
  #endif /* UL_WITH_DEVFS */

  printk(KERN_INFO "ul_usb_init_chan: minor=%d baud=%d my_adr=%d ready\n",
	 minor,udrv->baud_val,udrv->my_adr);
  
  if(minor>=0) ul_drv_arr[minor]=udrv;
  *pudrv=udrv;
  return 0;
}

#if (LINUX_VERSION_CODE < VERSION(2,5,41))
static void *ul_usb_probe(struct usb_device *dev, unsigned int ifnum, const struct usb_device_id *id)
{
  unsigned long driver_info=id->driver_info;
  ul_drv *udrv;
  printk(KERN_INFO "ulan_init_one: USB device found devnum=%d, ifnum=%d\n",
         dev->devnum, ifnum);
  switch(driver_info&~0xff){
    case  UL_USB_HW_PS1:
      if(ul_usb_init_chan(dev,ifnum,id,&udrv,ul_usb_ps1_init,driver_info)<0)
        return NULL;
      printk(KERN_INFO "ul_usb_probe: data type\n");
      return udrv;
  }
  printk(KERN_CRIT "ul_usb_probe: No device of specified driver_data type\n");
  return NULL;
}
#else /* 2.5.41 */
static int ul_usb_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
  struct usb_device *dev = interface_to_usbdev(intf);
  int ifnum = intf->altsetting->desc.bInterfaceNumber;
  unsigned long driver_info=id->driver_info;
  ul_drv *udrv;
  dev_set_drvdata(&intf->dev,NULL);
  printk(KERN_INFO "ulan_init_one: USB device found devnum=%d, ifnum=%d\n",
         dev->devnum, ifnum);
  switch(driver_info&~0xff){
    case  UL_USB_HW_PS1:
      if(ul_usb_init_chan(dev,ifnum,id,&udrv,ul_usb_ps1_init,driver_info)<0)
        return -EIO;
      printk(KERN_INFO "ul_usb_probe: data type\n");
      dev_set_drvdata(&intf->dev,udrv);
      return 0;
  }
  printk(KERN_CRIT "ul_usb_probe: No device of specified driver_data type\n");
  return -EIO;
}
#endif /* 2.5.41 */

#if (LINUX_VERSION_CODE < VERSION(2,5,41))
static void ul_usb_disconnect(struct usb_device *dev, void *ptr)
{
  ul_drv *udrv, *next_udrv;
  int i;
  udrv=(ul_drv *)ptr;
  if(!udrv){
    printk(KERN_CRIT "ulan_remove_one: no uLan drvdata\n");
    return;
  }
  for(;udrv;udrv=next_udrv){
    if(udrv->magic!=UL_DRV_MAGIC){
      printk(KERN_CRIT "ulan_remove_one: Wrong uLan MAGIC number!!!\n");
      return;
    }
    next_udrv=udrv->next_chan;
    if(dev!=(struct usb_device *)udrv->dev){
      printk(KERN_CRIT "ul_usb_disconnect: BAD - cross USB device remove\n");
    }
    for(i=0;i<UL_MINORS;i++){
      if (udrv==ul_drv_arr[i]) ul_drv_arr[i]=NULL;
    }
    #ifdef UL_WITH_DEVFS
    if(udrv->devfs_handle) kc_devfs_delete(udrv->devfs_handle);
    #endif /* UL_WITH_DEVFS */
    ul_drv_free(udrv);
  }
  printk(KERN_INFO "ul_usb_disconnect: USB device removed\n");
}
#else /* 2.5.41 */
static void ul_usb_disconnect(struct usb_interface *intf)
{
  struct usb_device *dev = interface_to_usbdev(intf);
  ul_drv *udrv, *next_udrv;
  int i;
  udrv=(ul_drv *)dev_get_drvdata(&intf->dev);
  dev_set_drvdata(&intf->dev,NULL);
  if(!udrv){
    printk(KERN_CRIT "ulan_remove_one: no uLan drvdata\n");
    return;
  }
  for(;udrv;udrv=next_udrv){
    if(udrv->magic!=UL_DRV_MAGIC){
      printk(KERN_CRIT "ulan_remove_one: Wrong uLan MAGIC number!!!\n");
      return;
    }
    next_udrv=udrv->next_chan;
    if(dev!=(struct usb_device *)udrv->dev){
      printk(KERN_CRIT "ul_usb_disconnect: BAD - cross USB device remove\n");
    }
    for(i=0;i<UL_MINORS;i++){
      if (udrv==ul_drv_arr[i]) ul_drv_arr[i]=NULL;
    }
    #ifdef UL_WITH_DEVFS
    if(udrv->devfs_handle) kc_devfs_delete(udrv->devfs_handle);
    #endif /* UL_WITH_DEVFS */
    ul_drv_free(udrv);
  }
  printk(KERN_INFO "ul_usb_disconnect: USB device removed\n");
}
#endif /* 2.5.41 */

/*******************************************************************/
/* usb2ulan device requests */

#define U2U_PKT_BUF_SIZE       32     /* driver maximal supported packed size */
#define U2U_MAX_MSG_INPR        4     /* maximal number of messages in progress */

#define VENDOR_START_ULAN       0     /* Inform converter about uLan driver ready */
#define VENDOR_STOP_ULAN        1     /* Inform about uLan driver stopping  */
#define VENDOR_IS_RUNNING_ULAN  2
#define VENDOR_ARC_ON_ULAN      3     /* Set promiscuous monitoring mode */
#define VENDOR_ARC_OFF_ULAN     4     /* Return to std. adress recognition mode */

#define U2UBFL_NORE     0x040         /* Do not try to repeat if error occurs */ 
#define U2UBFL_TAIL     0x020         /* Message has tail frame */
#define U2UBFL_REC      0x010         /* Request receiption of block */
#define U2UBFL_FAIL     0x008         /* Message cannot be send - error */
#define U2UBFL_PROC	0x004         /* Message succesfull send */
#define U2UBFL_AAP	0x003         /* Request imediate proccessing of frame by receiver station with acknowledge */
#define U2UBFL_PRQ	0x002         /* Request imediate proccessing of frame by receiver station */
#define U2UBFL_ARQ	0x001         /* Request imediate acknowledge by receiving station */

/* Definition of chip_buff usage */
#define ul_usb_in_urb_st chip_buff[0] /* completion status of  ?_complete_in */
#define ul_usb_max_pkt_size chip_buff[1] /* maximal USB packet size */
#define ul_usb_max_msg_inpr chip_buff[2] /* maximal number of messages in progress */


#if (LINUX_VERSION_CODE <= VERSION(2,4,10))
  #define usb_fill_bulk_urb FILL_BULK_URB
#endif

#if (LINUX_VERSION_CODE < VERSION(2,5,0))
  #define kc_usb_alloc_urb(urb, gfp) usb_alloc_urb(urb) 
  #define kc_usb_submit_urb(urb, gfp) usb_submit_urb(urb) 
#else
  #define kc_usb_alloc_urb usb_alloc_urb
  #define kc_usb_submit_urb usb_submit_urb
#endif

/* Support for multiple submitted messages */

typedef struct ul_usb_tx_fifo_slot{
  ul_mem_blk    *msg;
  unsigned char ustamp;
}ul_usb_tx_fifo_slot;

typedef struct ul_usb_tx_fifo{
  ul_usb_tx_fifo_slot slot[U2U_MAX_MSG_INPR];
  int size;
  int used;
  int idxin;
  int idxout;
}ul_usb_tx_fifo;

static int ul_usb_tx_fifo_empty(const ul_usb_tx_fifo *fifo)
{
  return fifo->slot[fifo->idxout].msg==NULL;
}

static int ul_usb_tx_fifo_full(const ul_usb_tx_fifo *fifo)
{
  return fifo->slot[fifo->idxin].msg!=NULL;
}

static int ul_usb_tx_fifo_incidx(const ul_usb_tx_fifo *fifo, int idx)
{
  if(++idx>=fifo->size) idx=0;
  return idx;
}

static void ul_usb_tx_fifo_print(const ul_usb_tx_fifo *fifo)
{
  int i;
  UL_PRINTF(KERN_CRIT "ul_usb: tx_fifo: idxin=%d idxout=%d [",
            fifo->idxin,fifo->idxout);
  for(i=0;i<fifo->size;i++){
    if(i) UL_PRINTF(",");
    if(!fifo->slot[i].msg)
      UL_PRINTF("Empty");
    else
      UL_PRINTF("%d",fifo->slot[i].ustamp);
  }
  UL_PRINTF("]\n");
}

/*** Input completion for Petr Smolik's USB2uLan converter ***/
void ul_usb_ps1_complete_in(struct urb *urb)
{
  ul_drv *udrv;
  int st;
  udrv=urb->context;
  if(udrv->magic!=UL_DRV_MAGIC){
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_complete_in: Wrong uLan MAGIC number!!!\n");
    return;
  }
  if(urb->status<0) 
    st=16-urb->status;
   else st=2;
  LOG_IRQ(KERN_CRIT "ul_usb_ps1_complete_in: status %d\n",urb->status);
  UL_MB();
  udrv->ul_usb_in_urb_st=st;
  uld_atomic_set_dfl(udrv,ASK_ISR);
  wake_up(&udrv->wq_kwt);
}


/*** Send one frame to USB2uLan converter ***/
int ul_usb_ps1_send_frame(ul_drv *udrv, ul_mem_blk *frame,
                           unsigned char ustamp)
{
  unsigned char buf[U2U_PKT_BUF_SIZE];
  int i, len, ret;
  ul_data_it di;
  struct usb_device *dev=(struct usb_device *)udrv->dev;
  int max_pkt=udrv->ul_usb_max_pkt_size;

  buf[0]=UL_BLK_HEAD(frame).dadr;
  buf[1]=UL_BLK_HEAD(frame).sadr;
  buf[2]=UL_BLK_HEAD(frame).cmd;
  i=UL_BLK_HEAD(frame).flg; buf[3]=0;
  if(i&UL_BFL_ARQ) buf[3]|=U2UBFL_ARQ;
  if(i&UL_BFL_PRQ) buf[3]|=U2UBFL_PRQ;
  if(i&UL_BFL_NORE) buf[3]|=U2UBFL_NORE;
  if(i&UL_BFL_TAIL) buf[3]|=U2UBFL_TAIL;
  if(i&UL_BFL_REC) buf[3]|=U2UBFL_REC;
  buf[4]=ustamp;
  buf[5]=0;
  len=UL_BLK_HEAD(frame).len;
  if(i&UL_BFL_REC){	/* Placeholder message for receive does not */
    len=0;		/* need data, but should check UL_BFL_LNMM !!! */
  }
  buf[6]=len; buf[7]=len/0x100;

  LOG_IRQ(KERN_CRIT "ul_usb : sending head URB (d:%d,s:%d,c:%X,f:%X,l:%d) stmp %d\n",
               buf[0],buf[1],buf[2],buf[3],
               buf[6]+0x100*buf[7],buf[4]);

  ret=usb_bulk_msg(dev, usb_sndbulkpipe(dev, 1), buf, 8, &i, ULD_HZ);
  if(ret<0) {
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_send_frame: usb_bulk_write head error %d returned\n",ret);
    return ret;
  }

  ul_di_init(&di,frame);
  while(len>0) {
    i=len<max_pkt?len:max_pkt; len-=i;
    ul_di_read(&di, buf, i);
    ret=usb_bulk_msg(dev, usb_sndbulkpipe(dev, 1), buf, i, &i, ULD_HZ);
    if(ret<0) {
      LOG_FATAL(KERN_CRIT "ul_usb_ps1_send_frame: usb_bulk_write data error %d returned\n",ret);
      return ret;
    }
  }
  return di.pos;
}

/*** Receive data of one frame from USB2uLan converter ***/
int ul_usb_ps1_receive_data(ul_drv *udrv, ul_mem_blk *frame, int len)
{
  unsigned char buf[U2U_PKT_BUF_SIZE];
  int i, ret;
  ul_data_it di;
  struct usb_device *dev=(struct usb_device *)udrv->dev;
  int max_pkt=udrv->ul_usb_max_pkt_size;
  
  ul_di_init(&di,frame);
  while(len>0) {
    i=len<max_pkt?len:max_pkt; len-=i;
    ret=usb_bulk_msg(dev, usb_rcvbulkpipe(dev, 2), buf, i, &i, ULD_HZ);
    if(ret<0) {
      LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: usb_bulk_read data error %d returned\n",ret);
      return ret;
      break;
    } else {
      if(i!=ul_di_write(&di, buf, i)){
	return -EIO;
      }
    }
  }
  return di.pos;
}


/*** Worker thread for Petr Smolik's USB2uLan converter ***/
int ul_usb_ps1_thread(void *ptr)
{
  ul_drv *udrv=(ul_drv *)ptr;
  struct usb_device *dev;
  int ret;
  int i, len;
  int fatal_err=0;
  volatile unsigned char in_buf[8];
  unsigned char u2u_ret[2];
  unsigned char ustamp=0;
  ul_mem_blk *tx_mes=NULL;
  ul_mem_blk *rx_mes=NULL;
  ul_mem_blk *frame;
  ul_mem_blk *tail_frame=NULL;
  long tx_sta_time;
  ul_usb_tx_fifo tx_fifo;
  struct urb *in_urb;
  volatile unsigned in_urb_st;

  if(udrv->magic!=UL_DRV_MAGIC)
  {
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: Wrong uLan MAGIC number!!!\n");
    uld_atomic_clear_dfl(udrv,KWTRUN);
    return -ENODEV;
  }
  dev=(struct usb_device *)udrv->dev;
  if(udrv->ul_usb_max_pkt_size>U2U_PKT_BUF_SIZE)
    udrv->ul_usb_max_pkt_size=U2U_PKT_BUF_SIZE;

  if(udrv->ul_usb_max_msg_inpr>U2U_MAX_MSG_INPR)
    udrv->ul_usb_max_msg_inpr=U2U_MAX_MSG_INPR;
  memset(&tx_fifo,0,sizeof(tx_fifo));
  tx_fifo.size=udrv->ul_usb_max_msg_inpr;
  
#if (LINUX_VERSION_CODE >= VERSION(2,6,0))
  daemonize("kulusbps1d");
  /*allow_signal(SIGKILL);*/
#else /* <2.6.0 */
  lock_kernel();
  daemonize();

 #if (LINUX_VERSION_CODE >= VERSION(2,4,10))
  exit_files(current);
  current->files = init_task.files;
  atomic_inc(&current->files->count);
  daemonize();
  reparent_to_init();
 #endif /* >=2.4.10 */

  /* avoid getting signals */
  spin_lock_irq(&current->sigmask_lock);
  flush_signals(current);
  sigfillset(&current->blocked);
  recalc_sigpending(current);
  spin_unlock_irq(&current->sigmask_lock);

  set_fs(KERNEL_DS);
  strcpy(current->comm, "kulusbps1d");
  unlock_kernel();
#endif /* <2.6.0 */

  if(udrv->baud_val)
    udrv->baud_div=(udrv->baud_base+udrv->baud_val/2)/udrv->baud_val;

  /* Set converter parameters */
  ret=usb_control_msg(dev, usb_rcvctrlpipe(dev,0), 
                      VENDOR_START_ULAN, USB_TYPE_VENDOR,
                      /*value*/ udrv->baud_val, /*index*/ udrv->my_adr, u2u_ret, 1, 2*ULD_HZ);
  
  if(ret<0) {
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: KWT usb_control_msg error %d returned\n",ret);
    fatal_err=1;
  }

  if(!u2u_ret[0]) {
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: not supported baudrate (%d)\n",udrv->baud_val);
    fatal_err=1;
  }
 
  if((in_urb=kc_usb_alloc_urb(0, GFP_KERNEL))) {
    /* URB allocated OK */
  }else{
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: KWT cannot alloc in_urb\n");
    fatal_err=1;
  }
  usb_clear_halt(dev, usb_sndbulkpipe(dev, 1));
  usb_clear_halt(dev, usb_rcvbulkpipe(dev, 2));
  udrv->ul_usb_in_urb_st=0;
 
  while(!uld_atomic_test_dfl(udrv,KWTKILL) && !fatal_err){
    uld_atomic_clear_dfl(udrv,ASK_ISR);
    
    while(!ul_usb_tx_fifo_full(&tx_fifo)){ /* when free Tx slot */

      /* Initiate sending of message from the prep_bll queue */
      if(!(tx_mes=udrv->con_message=udrv->prep_bll.first)) break;

      /* printudrvbll(udrv); */
      ul_bll_move_mes(&udrv->work_bll,tx_mes);
      LOG_MESSAGES(KERN_INFO "ul_usb: KWT processing message\n");
      
      if(!++ustamp) ustamp=1;

      /* send frames of message */
      for(frame=tx_mes;frame;frame=UL_BLK_HEAD(frame).next){
	ret=ul_usb_ps1_send_frame(udrv, frame, ustamp);
	if(ret<0) {
          usb_clear_halt(dev, usb_sndbulkpipe(dev, 1));
          UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
          ul_bll_move_mes(&udrv->proc_bll,tx_mes);
          SCHEDULE_BH(udrv);
          /* ulan_do_bh(udrv); */
	  tx_mes=udrv->con_message=NULL;
          break;
	}
	/* Message has no more frames */
	if(!(UL_BLK_HEAD(frame).flg&UL_BFL_TAIL)||
	   (UL_BLK_HEAD(frame).next==NULL)){
	  tx_fifo.slot[tx_fifo.idxin].ustamp=ustamp;
	  tx_fifo.slot[tx_fifo.idxin].msg=tx_mes;
          tx_fifo.idxin=ul_usb_tx_fifo_incidx(&tx_fifo,tx_fifo.idxin);
          tx_sta_time=jiffies;
          ul_usb_tx_fifo_print(&tx_fifo);
	}
      }
    }
    
    UL_MB();
    in_urb_st=*(volatile unsigned *)&(udrv->ul_usb_in_urb_st);
    UL_MB();
    if(in_urb_st==2) {
      UL_MB();
      LOG_IRQ(KERN_CRIT "ul_usb : received head URB (d:%d,s:%d,c:%X,f:%X,l:%d) stmp %d\n",
                   in_buf[0],in_buf[1],in_buf[2],in_buf[3],
                   in_buf[6]+0x100*in_buf[7],in_buf[4]);
      if(in_buf[3]&(U2UBFL_PROC|U2UBFL_FAIL)){
        tx_mes=NULL;
        ul_usb_tx_fifo_print(&tx_fifo);
        if(!ul_usb_tx_fifo_empty(&tx_fifo)){
          if(in_buf[4]==tx_fifo.slot[tx_fifo.idxout].ustamp){
	    tx_mes=tx_fifo.slot[tx_fifo.idxout].msg;
	  }else{
            i=ul_usb_tx_fifo_incidx(&tx_fifo,tx_fifo.idxout);
	    while((i!=tx_fifo.idxout)&&(tx_fifo.slot[i].msg!=NULL)){
	      if(in_buf[4]==tx_fifo.slot[i].ustamp){
	        do{
                  tx_mes=tx_fifo.slot[tx_fifo.idxout].msg;
	          UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
                  ul_bll_move_mes(&udrv->proc_bll,tx_mes);
		  tx_fifo.slot[tx_fifo.idxout].msg=NULL;
		  tx_fifo.idxout=ul_usb_tx_fifo_incidx(&tx_fifo,tx_fifo.idxout);
                }while(tx_fifo.idxout!=i);
                SCHEDULE_BH(udrv);
	        tx_mes=tx_fifo.slot[tx_fifo.idxout].msg;
              }
              i=ul_usb_tx_fifo_incidx(&tx_fifo,i);
            }
	    if(!tx_mes){
              LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: incorrect ustamp of returned message\n");
	    }else{
              LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: some lost ustamp messages\n");
	    }
          }
	}else{
	  LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: tx_mes for returned message\n");
	}

	if(!tx_mes){
          usb_clear_halt(dev, usb_rcvbulkpipe(dev, 2));
	}else{
	  if(tail_frame&&(UL_BLK_HEAD(tx_mes).flg&UL_BFL_TAIL)){
	    frame=tail_frame;
	    LOG_MESSAGES("ul_usb_ps1_thread: tail receiption\n");
	  }else{
	    frame=tx_mes;
	  }

          if(in_buf[3]&U2UBFL_FAIL){
	    UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
	  }else if(in_buf[3]&U2UBFL_REC) {
	    if(!(UL_BLK_HEAD(frame).flg&UL_BFL_REC)){
	      UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
              LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: tail rec mitchmatch\n");
	    }else{
	      len=in_buf[6]+in_buf[7]*0x100;
	      UL_BLK_HEAD(frame).dadr=in_buf[0];
	      UL_BLK_HEAD(frame).sadr=in_buf[1];
	      UL_BLK_HEAD(frame).cmd=in_buf[2];
	      UL_BLK_HEAD(frame).len=len;
              ret=ul_usb_ps1_receive_data(udrv, frame, len);
	      if(ret<0) {
		UL_BLK_HEAD(tx_mes).flg|=UL_BFL_FAIL;
        	LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: tail receive problem\n");
	      }
	    }
	  }
	    
	  if((in_buf[3]&U2UBFL_TAIL)&&(UL_BLK_HEAD(frame).flg&UL_BFL_TAIL)
	      &&!(UL_BLK_HEAD(tx_mes).flg&UL_BFL_FAIL)){
	    tail_frame=UL_BLK_HEAD(frame).next;
	    LOG_MESSAGES("ul_usb_ps1_thread: preparing for tail receive\n");
          }else{
	    if(UL_BLK_HEAD(tx_mes).flg&UL_BFL_M2IN || UL_BLK_HEAD(tx_mes).flg&UL_BFL_FAIL ||
	     UL_BLK_HEAD(tx_mes).dadr==udrv->my_adr || !UL_BLK_HEAD(tx_mes).dadr){
              ul_bll_move_mes(&udrv->proc_bll,tx_mes);
              SCHEDULE_BH(udrv);
              /* ulan_do_bh(udrv); */
	    }else{
              ul_bll_free_mes(tx_mes);
	    }
	    tx_mes=NULL;
	    tail_frame=NULL;
	    tx_fifo.slot[tx_fifo.idxout].msg=NULL;
	    tx_fifo.idxout=ul_usb_tx_fifo_incidx(&tx_fifo,tx_fifo.idxout);
	    ul_usb_tx_fifo_print(&tx_fifo);
	  }
	}
      } else {
	LOG_MESSAGES("uLan : begin of receive message\n");
	i=0;
        /* if(in_buf[3]&U2UBFL_TAIL ) i|=UL_BFL_TAIL; */
        /* if(in_buf[3]&U2UBFL_REC) i|=UL_BFL_REC; */
        rx_mes=ul_new_frame_head(udrv,in_buf[0]/*dadr*/,in_buf[1]/*sadr*/, 
             	                 in_buf[2]/*cmd*/, i/*flg*/);
	if(rx_mes==NULL){ 
	  LOG_FAILS(KERN_ERR "uLan : NO memory for receive\n");
          usb_clear_halt(dev, usb_rcvbulkpipe(dev, 2));
	}else{
	  ul_bll_ins(&udrv->work_bll,rx_mes);
	  len=in_buf[6]+in_buf[7]*0x100;

          ret=ul_usb_ps1_receive_data(udrv, rx_mes, len);

	  if(ret<0) {
            LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: message receive problem\n");
            ul_bll_free_mes(rx_mes);
	  }else{
	    UL_BLK_HEAD(rx_mes).len=len;
            UL_BLK_HEAD(rx_mes).stamp=ul_gen_stamp();
            ul_bll_move_mes(&udrv->proc_bll,rx_mes);
	  }

          rx_mes=NULL;
          SCHEDULE_BH(udrv);
          /* ulan_do_bh(udrv); */
          /* printudrvbll(udrv); */
	}
      }
    }
    
    if(in_urb_st!=1)
    {
      ret=16-udrv->ul_usb_in_urb_st;
      if((ret==-ENODEV)||(ret==-EILSEQ)||(ret==-EPROTO)){
        LOG_FATAL(KERN_INFO "ul_usb: usb2ulan device problem detected\n");
        fatal_err=1;
        continue;
      }
      *(volatile unsigned *)&(udrv->ul_usb_in_urb_st)=1;
      UL_MB();
      if(uld_atomic_test_dfl(udrv,KWTKILL)) break;
      LOG_IRQ(KERN_INFO "ul_usb: submitting URB\n");
      usb_fill_bulk_urb (in_urb, dev, usb_rcvbulkpipe(dev, 2),
          in_buf, 8, ul_usb_ps1_complete_in, udrv);
      ret=kc_usb_submit_urb(in_urb, GFP_KERNEL);
      if (ret==-EPIPE) {
        usb_clear_halt(dev, usb_rcvbulkpipe(dev, 2));
        ret=kc_usb_submit_urb(in_urb, GFP_KERNEL);
      }  
      if(ret<0) {
        LOG_FATAL(KERN_INFO "ul_usb: KWT usb_submit_urb fatal error %d\n",ret);
        fatal_err=1;
      }
    } else {
      LOG_IRQ(KERN_INFO "ul_usb: KWT sleeping\n");
      wait_event(udrv->wq_kwt, uld_test_dfl(udrv,ASK_ISR));
      LOG_IRQ(KERN_INFO "ul_usb: KWT wakeup\n");
    }
  }
  /* Set converter parameters */
  ret=usb_control_msg(dev, usb_rcvctrlpipe(dev,0),
                      VENDOR_STOP_ULAN, USB_TYPE_VENDOR,
                      /*value*/ 0, /*index*/ 0, NULL, 0, ULD_HZ);

  if(ret<0) {
    LOG_FATAL(KERN_CRIT "ul_usb_ps1_thread: KWT usb_control_msg error %d returned\n",ret);
    fatal_err=1;
  }
  
  ul_usb_tx_fifo_print(&tx_fifo);
  while(!ul_usb_tx_fifo_empty(&tx_fifo)){
    ul_bll_free_mes(tx_fifo.slot[tx_fifo.idxout].msg);
    tx_fifo.slot[tx_fifo.idxout].msg=NULL;
    tx_fifo.idxout=ul_usb_tx_fifo_incidx(&tx_fifo,tx_fifo.idxout);
  }
  
  if(in_urb) {
    usb_unlink_urb(in_urb);
    usb_free_urb(in_urb);
  }
  
  lock_kernel();
  uld_atomic_clear_dfl(udrv,KWTRUN);
  /*complete_and_exit(&unload_sem, 0);*/
  wake_up(&udrv->wq_kwt);
  unlock_kernel();
  return 0;
}

/*** Wake up worker thread  ***/
int ul_usb_ps1_stroke(ul_drv *udrv)
{
  uld_atomic_set_dfl(udrv,ASK_ISR);
  wake_up(&udrv->wq_kwt);
  return 0;
}

/*** Control functions of chip driver  ***/
int ul_usb_ps1_cctrl(ul_drv *udrv, int ctrl_fnc, int param)
{
  switch (ctrl_fnc)
  {
   case UL_CC_PINIT:	/* initialize ports (parameter baud)*/
  	{ return 0;
  	};
   case UL_CC_PDONE:	/* deinitialize ports */
  	{ return 0;
  	};
   case UL_CC_RQIRQ:	/* request IRQ handling - kernel worker thread*/
  	{ if(uld_test_and_set_dfl(udrv,KWTRUN))
          { LOG_IRQ("ul_usb_ps1_cctrl: KWT already running\n");
            return 0;
          };
          uld_atomic_clear_dfl(udrv,KWTKILL);    
          if(kernel_thread(ul_usb_ps1_thread, udrv,
           #if (LINUX_VERSION_CODE < VERSION(2,5,41))
            CLONE_FS | CLONE_FILES | CLONE_SIGNAL
           #else /* 2.5.41 */
            CLONE_KERNEL
           #endif /* 2.5.41 */
          )<0){
	    LOG_FATAL(KERN_CRIT "ul_usb_ps1_cctrl: Cannot fork KWT thread!!!\n");
	    uld_atomic_clear_dfl(udrv,KWTRUN);
	    return UL_RC_EKWT;
	  }
	  LOG_IRQ("ul_usb_ps1_cctrl: KWT forked\n");
	  return 0;
  	};
   case UL_CC_FREEIRQ:	/* release IRQ handling - kernel worker thread*/
  	{ 
	  if(!uld_atomic_test_dfl(udrv,KWTRUN))
          { LOG_IRQ("ul_usb_ps1_cctrl: KWT already stopped\n");
            return 0;
          };
	  uld_atomic_set_dfl(udrv,KWTKILL);
          uld_atomic_set_dfl(udrv,ASK_ISR);
	  wake_up(&udrv->wq_kwt);
	  wait_event(udrv->wq_kwt, !uld_test_dfl(udrv,KWTRUN));
	  LOG_IRQ("ul_usb_ps1_cctrl: KWT stopped\n");
	  return 0;
  	};
  };
  return UL_RC_ENOFNC;
};

/*** Petr Smolik's USB2ULAN version 1 initialize ***/
int ul_usb_ps1_init(ul_drv *udrv, int port, int irq, int baud, int options)
{
  udrv->chip_options=options;
  if(!baud) baud=19200;
  udrv->baud_val=baud;
  udrv->baud_base=57600;
  udrv->baud_div=(udrv->baud_base+baud/2)/baud;
  udrv->port=port;
  udrv->irq=0; 
  udrv->fnc_recch=NULL;
  udrv->fnc_sndch=NULL;
  udrv->fnc_wait=NULL;
  udrv->fnc_pool=NULL;
  udrv->fnc_connect=NULL;
  udrv->fnc_cctrl=&ul_usb_ps1_cctrl;
  udrv->fnc_stroke=&ul_usb_ps1_stroke;
  udrv->ul_usb_max_pkt_size=8; /* <= U2U_PKT_BUF_SIZE */
  udrv->ul_usb_max_msg_inpr=ul_usb_msg_inpr;
  if(!udrv->ul_usb_max_msg_inpr)
    udrv->ul_usb_max_msg_inpr=1; /* <= U2U_MAX_MSG_INPR */
  return 0;
};


/*******************************************************************/

/*** initialize kernel module ***/
int init_module(void)
{
 #ifdef UL_WITH_DEVFS
  kc_devfs_handle_t devfs_handle;
  char devname[32];
 #endif /* UL_WITH_DEVFS */
  ul_drv* udrv;
  int ret;
  int minor;

  printk(KERN_INFO "uLan v" UL_DRV_VERSION " init_module\n");
  printk(KERN_INFO "Compiled at %s %s\n",__DATE__, __TIME__);

  if(debug!=-1){
    uld_debug_flg=debug;
  }

  ret=0;
 #ifdef UL_WITH_UARTS
  for(minor=0;minor<UL_MINORS;minor++)
  {
    if(port[minor]||chip[minor])
    { /* get new ul_drv (port, irq, baud, my_adr); */
      udrv=ul_drv_new(port[minor],irq[minor],
   		      baud[minor],my_adr[minor],chip[minor]);
      ul_drv_arr[minor]=udrv;
      if(udrv){
	#ifdef UL_WITH_DEVFS
	  sprintf (devname, "ulan%d", minor);
          devfs_handle=kc_devfs_new_cdev(NULL, MKDEV(ulan_major_dev, minor), 
			S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, 
			&ulan_fops, udrv, devname);
	  udrv->devfs_handle=devfs_handle;
	#endif /* UL_WITH_DEVFS */
	ret++;
      };
    };
  };
 #endif /*UL_WITH_UARTS*/

 #ifdef UL_WITH_PCI
  { int rc;
    rc=pci_register_driver (&ulan_pci_driver);
    ret+=rc;
  }
 #endif  /* UL_WITH_PCI */

 #ifdef UL_WITH_USB
  /* register this driver with the USB subsystem */
  { int rc;
    rc = usb_register(&ul_usb_driver);
    if (rc < 0) {
      printk(KERN_CRIT "usb_register failed for the "__FILE__" driver. Error number %d",
	 rc);
    } else {
      ret++;
    }
  }
 #endif  /* UL_WITH_USB */


  if(!ret)
  { printk(KERN_ERR "uLan module_init : no hardware initialized, aborting !!!\n");
    #ifdef UL_WITH_USB
     usb_deregister(&ul_usb_driver);
    #endif  /* UL_WITH_USB */
    #ifdef UL_WITH_PCI
    pci_unregister_driver (&ulan_pci_driver);
    #endif  /* UL_WITH_PCI */
    return -ENODEV;
  };

 #ifdef UL_WITH_DEVFS
  ret = devfs_register_chrdev(ulan_major_dev, "ulan", &ulan_fops);
 #else /* UL_WITH_DEVFS */
  ret = register_chrdev(ulan_major_dev, "ulan", &ulan_fops);
 #endif /* UL_WITH_DEVFS */
  printk(KERN_INFO "uLan init : called register dev %d and got %d\n",
			 ulan_major_dev,ret);
  if (ret < 0) {
    printk(KERN_ERR "uLan init : unable to grab device #\n");
    ulan_major_dev=0;
    #ifdef UL_WITH_PCI
    pci_unregister_driver (&ulan_pci_driver);
    #endif  /* UL_WITH_PCI */
    for(minor=0;minor<UL_MINORS;minor++)
    { udrv=ul_drv_arr[minor];
      if(udrv)
      { ul_drv_arr[minor]=NULL;
	#ifdef UL_WITH_DEVFS
	if(udrv->devfs_handle) kc_devfs_delete(udrv->devfs_handle);
	#endif /* UL_WITH_DEVFS */
	ul_drv_free(udrv);
      };
    };
    return ret;
  }
  if(!ulan_major_dev) ulan_major_dev=ret;
  printk(KERN_INFO "uLan init : driver succesfully registered as major %d\n",
	          ulan_major_dev);

  #ifdef UL_INT_TST
  udrv=ul_drv_arr[0];
  if(udrv)
  { printk("uLan : geninfoblk\n");
    geninfoblk(udrv);
    printk("uLan : gentestblk\n");
    gentestblk(udrv);
    printk("uLan : uld_timeout\n");
    uld_timeout(udrv);
  };
  #endif /* UL_INT_TST */

  return 0;
}

/*** cleanup kernel module ***/
void cleanup_module(void)
{
  ul_drv* udrv;
  int i;
  int ret;

  if (ulan_major_dev != 0) {
   #ifdef UL_WITH_DEVFS
    ret = devfs_unregister_chrdev(ulan_major_dev, "ulan");
   #else /* UL_WITH_DEVFS */
    ret = unregister_chrdev(ulan_major_dev, "ulan");
   #endif /* UL_WITH_DEVFS */
    if(ret>=0)
      printk(KERN_INFO "uLan cleanup: Unregister  O.K.\n");
    else
      printk(KERN_CRIT "uLan cleanup: Unregister failed\n");
  };

  udrv=ul_drv_arr[0];
  if(udrv)
  { printk("uLan : printudrvbll\n");
    printudrvbll(udrv);
    printk("uLan : max cycle cnt  %d\n",ul_max_cycle_cnt);
    printk("uLan : max cycle time %d\n",ul_max_cycle_time);
  }

 #ifdef UL_WITH_USB
  usb_deregister(&ul_usb_driver);
 #endif  /* UL_WITH_USB */
 #ifdef UL_WITH_PCI
  pci_unregister_driver (&ulan_pci_driver);
 #endif  /* UL_WITH_PCI */
  for(i=0;i<UL_MINORS;i++)
  { udrv=ul_drv_arr[i];
    ul_drv_arr[i]=NULL;
    if(udrv){
     #ifdef UL_WITH_DEVFS
     if(udrv->devfs_handle) kc_devfs_delete(udrv->devfs_handle);
     #endif /* UL_WITH_DEVFS */
     ul_drv_free(udrv);
    }
  }
 #ifdef ENABLE_UL_MEM_CHECK
  UL_PRINTF("MEM_CHECK: malloc-free=%d\n",
		 (int)atomic_read(&ul_mem_check_counter));
 #endif /* ENABLE_UL_MEM_CHECK */
}

#endif /* FOR_LINUX_KERNEL */

